# -----------------------------------------------------------------------------
#  pytermor [ANSI formatted terminal output toolset]
#  (c) 2022-2023. A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
"""
Package containing a set of formatters for prettier output, as well as utility classes
for removing some of the boilerplate code when dealing with escape sequences. Also
includes several Python Standard Library methods rewritten for correct work with
strings containing control sequences.
"""
from __future__ import annotations

import codecs
import math
import os
import re
import textwrap
import typing as t
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field
from functools import reduce
from math import ceil
from typing import Union

from .common import ArgTypeError, Align
from .utilmisc import chunk, get_terminal_width

_PRIVATE_REPLACER = "\U000E5750"


def pad(n: int) -> str:
    return " " * n


def padv(n: int) -> str:
    return "\n" * n


def ljust_sgr(s: str, width: int, fillchar: str = " ", actual_len: int = None) -> str:
    """
    SGR-formatting-aware implementation of ``str.ljust``.

    Return a left-justified string of length ``width``. Padding is done
    using the specified fill character (default is a space).
    """
    if actual_len is None:
        actual_len = len(SgrStringReplacer().apply(s))
    return s + fillchar * max(0, width - actual_len)


def rjust_sgr(s: str, width: int, fillchar: str = " ", actual_len: int = None) -> str:
    """
    SGR-formatting-aware implementation of ``str.rjust``.

    Return a right-justified string of length ``width``. Padding is done
    using the specified fill character (default is a space).
    """
    if actual_len is None:
        actual_len = len(SgrStringReplacer().apply(s))
    return fillchar * max(0, width - actual_len) + s


def center_sgr(s: str, width: int, fillchar: str = " ", actual_len: int = None) -> str:
    """
    SGR-formatting-aware implementation of ``str.center``.

    Return a centered string of length ``width``. Padding is done using the
    specified fill character (default is a space).

    .. todo ::

        поверить корректность работы в случае эмодзи (напр. 🔋)
        если алгоритм поедет -- можно заменить на f-стринги
    """
    if actual_len is None:
        actual_len = len(SgrStringReplacer().apply(s))

    fill_len = max(0, width - actual_len)
    if fill_len == 0:
        return s

    if actual_len % 2 == 1:
        right_fill_len = math.ceil(fill_len / 2)
    else:
        right_fill_len = math.floor(fill_len / 2)
    left_fill_len = fill_len - right_fill_len
    return (fillchar * left_fill_len) + s + (fillchar * right_fill_len)


def wrap_sgr(
    raw_input: str | list[str], width: int, indent_first: int = 0, indent_subseq: int = 0
) -> str:
    """
    A workaround to make standard library ``textwrap.wrap()`` more friendly
    to an SGR-formatted strings.

    The main idea is

    :param raw_input:
    :param width:
    """
    # initially was written as a part of es7s/core
    # package, and transferred here later
    sgrs: list[str] = []

    def push(m: t.Match):
        sgrs.append(m.group())
        return _PRIVATE_REPLACER

    if isinstance(raw_input, str):  # input can be just one paragraph
        raw_input = [raw_input]

    input = "\n\n".join(raw_input).split("\n\n")
    result = ""
    for raw_line in input:
        # had an inspiration and wrote this; no idea how does it work exactly, it just does
        replaced_line = re.sub(r"(\s?\S?)((\x1b\[([0-9;]*)m)+)", push, raw_line)
        wrapped_line = f"\n".join(
            textwrap.wrap(
                replaced_line,
                width=width,
                initial_indent=(indent_first * " "),
                subsequent_indent=(indent_subseq * " "),
            )
        )
        final_line = re.sub(_PRIVATE_REPLACER, lambda _: sgrs.pop(0), wrapped_line)
        result += final_line + "\n"
    return result


codecs.register_error("replace_with_qmark", lambda e: ("?", e.start + 1))


# =============================================================================
# Filters

ESCAPE_SEQ_REGEX = re.compile(
    r"""
    (?P<escape_char>\x1b)
    (?P<data>
      (?P<nf_class_seq>
        (?P<nf_interm>[\x20-\x2f]+)
        (?P<nf_final>[\x30-\x7e])
      )|
      (?P<fp_class_seq>
        (?P<fp_classifier>[\x30-\x3f])
        (?P<fp_params>[\x20-\x7e]*)
        (?P<fp_terminator>)
      )|
      (?P<fe_class_seq>
        (?P<fe_classifier>[\x40-\x5f])
        (?P<fe_params>[\x30-\x3f]*)
        (?P<fe_termintaor>[\x40-\x5a])
      )|
      (?P<fs_class_seq>
        (?P<fs_classifier>[\x60-\x7e])
        (?P<fs_params>[\x20-\x7e]*)
        (?P<fs_termintaor>)
      )  
    )
""",
    flags=re.VERBOSE,
)
# https://ecma-international.org/wp-content/uploads/ECMA-35_6th_edition_december_1994.pdf

SGR_SEQ_REGEX = re.compile(r"(\x1b)(\[)([0-9;]*)(m)")
CSI_SEQ_REGEX = re.compile(r"(\x1b)(\[)(([0-9;:<=>?])*)([@A-Za-z])")

CONTROL_CHARS = [*range(0x00, 0x08 + 1), *range(0x0E, 0x1F + 1), 0x7F]
WHITESPACE_CHARS = [*range(0x09, 0x0D + 1), 0x20]
PRINTABLE_CHARS = [*range(0x21, 0x7E + 1)]
NON_ASCII_CHARS = [*range(0x80, 0xFF + 1)]

ALIGN_FUNCS_SGR = {
    Align.LEFT: ljust_sgr,
    Align.CENTER: center_sgr,
    Align.RIGHT: rjust_sgr,
}
ALIGN_FUNCS_RAW = {
    Align.LEFT: str.ljust,
    Align.CENTER: str.center,
    Align.RIGHT: str.rjust,
}

IT = t.TypeVar("IT", str, bytes)  # input-type
OT = t.TypeVar("OT", str, bytes)  # output-type
PT = Union[IT, t.Pattern[IT]]  # pattern type
RT = Union[OT, t.Callable[[t.Match[OT]], OT]]  # replacer type
MT = t.Dict[int, IT]  # map
AT = Union["OmniFilter", t.Type["OmniFilter"]]

_dump_printers_cache: t.Dict[t.Type["AbstractTracer"], "AbstractTracer"] = dict()


class IFilter(t.Generic[IT, OT], metaclass=ABCMeta):
    """
    Main idea is to provide a common interface for string filtering, that can make
    possible working with filters like with objects rather than with functions/lambdas.
    """

    def __call__(self, s: IT) -> OT:
        """Can be used instead of `apply()`"""
        return self.apply(s)

    @abstractmethod
    def apply(self, inp: IT, extra: t.Any = None) -> OT:
        """
        Apply the filter to input *str* or *bytes*.

        :param inp:   input string
        :param extra: additional options
        :return: transformed string; the type can match the input type,
                 as well as be different -- that depends on filter type.
        """
        raise NotImplementedError


class NoopFilter(IFilter[IT, OT]):
    """ """

    def apply(self, inp: IT, extra: t.Any = None) -> OT:
        return inp


class OmniDecoder(IFilter[IT, str]):
    """ """

    def apply(self, inp: IT, extra: t.Any = None) -> str:
        return inp.decode() if isinstance(inp, bytes) else inp


class OmniEncoder(IFilter[IT, bytes]):
    """ """

    def apply(self, inp: IT, extra: t.Any = None) -> bytes:
        return inp.encode() if isinstance(inp, str) else inp


class StringAligner(IFilter[str, str]):
    def __init__(self, align: Align):
        self._align = align

    def apply(self, inp: str, raw_mode: bool = False) -> str:
        pass

    def _get_align_fn(self, raw_mode: bool = False):
        if raw_mode:
            return ALIGN_FUNCS_RAW.get(self._align)
        return ALIGN_FUNCS_SGR.get(self._align)


# -----------------------------------------------------------------------------
# Filters[Tracers]


class AbstractTracer(IFilter[IT, str], metaclass=ABCMeta):
    def __init__(self, char_per_line: int):
        self._char_per_line = char_per_line
        self._state: _TracerState = _TracerState()

    def apply(self, inp: IT, extra: TracerExtra = None) -> str:
        if len(inp) == 0:
            return "\n"

        self._state.reset(inp)
        self._state.inp_size_len = len(str(self._state.inp_size))
        self._state.offset_len = len(self._format_offset(self._state.inp_size))

        while len(inp) > 0:
            inp, part = inp[self._char_per_line :], inp[: self._char_per_line]
            self._process(part)
            self._state.lineno += 1
            self._state.offset += self._char_per_line

        header = self._format_line_separator("_", f"{extra.label}" if extra else "")
        footer = self._format_line_separator(
            "-", label_right="(" + self._format_offset(self._state.inp_size) + ")"
        )

        result = header + "\n"
        result += "\n".join(self._render_rows())
        result += "\n" + footer + "\n"
        return result

    def _format_offset(self, override: int = None) -> str:
        offset = override or self._state.offset
        return str(offset).rjust(self._state.inp_size_len)

    def _format_line_separator(
        self, fill: str, label_left: str = "", label_right: str = ""
    ) -> str:
        return (
            label_left
            + fill * (self._get_output_line_len() - len(label_left) - len(label_right))
            + label_right
        )

    def _render_rows(self) -> t.Iterable[str]:
        for row in self._state.rows:
            row_str = ""
            for col_idx, col_val in enumerate(row):
                row_str += col_val.rjust(self._state.cols_max_len[col_idx])
            yield row_str

    def _get_vert_sep_char(self):
        return "|"

    def _get_output_line_len(self) -> int:
        # useless before processing
        if self._state.cols_max_len is None:
            return 0
        return sum(l for l in self._state.cols_max_len)

    @abstractmethod
    def _process(self, part: IT) -> str:
        raise NotImplementedError

    @classmethod
    def estimate_line_len(cls, char_per_line: int, inp_size: int) -> int:
        raise NotImplementedError


class BytesTracer(AbstractTracer[bytes]):
    """
    str/bytes as byte hex codes, grouped by 4

    .. code-block:: hexdump
       :caption: Example output

        0000  0A 20 32 31 36 20 20 20  E2 94 82 20 20 75 70 6C  |a
        0010  20 20 20 20 20 20 20 20  20 20 20 20 20 20 20 20  |a
    """

    GROUP_SIZE = 4

    def __init__(self, char_per_line: int = 32):
        super().__init__(self.GROUP_SIZE * ceil(char_per_line / self.GROUP_SIZE))

    def _process(self, part: IT):
        self._state.add_row(self._make_row(part))

    def _make_row(self, part: IT) -> t.List[str]:
        return [
            " ",
            self._format_offset(),
            " " + self._get_vert_sep_char() + "  ",
            *self._format_main(part),
        ]

    def _format_offset(self, override: int = None) -> str:
        offset = override or self._state.offset
        size_len = 2 * ceil(self._state.inp_size_len / 2)
        return f"{offset:0{size_len}X}"

    def _format_main(self, part: bytes) -> t.Iterable[str]:
        for c in chunk(part, self.GROUP_SIZE):
            yield (" ".join([f"{b:02X}" for b in (*c,)])).ljust(3 * self.GROUP_SIZE + 1)

    @classmethod
    def estimate_line_len(cls, char_per_line: int, inp_size: int) -> int:
        space_len = 4
        sep_len = 1
        offset_len = len(str(inp_size))
        main_len = (3 * char_per_line * cls.GROUP_SIZE) + (
            char_per_line // cls.GROUP_SIZE
        )
        return space_len + sep_len + offset_len + main_len


class AbstractStringTracer(AbstractTracer[str], metaclass=ABCMeta):
    def __init__(self, char_per_line: int):
        self._output_filters: t.List[IFilter] = []
        super().__init__(char_per_line)

    def _format_output_text(self, text: str) -> str:
        return apply_filters(text, *self._output_filters).ljust(self._char_per_line)


class StringTracer(AbstractStringTracer):
    """
    str as byte hex codes (UTF-8), grouped by characters

    .. code-block:: hexdump
       :caption: Example output

       0056  45 4D 20 43 50 55     20     4F 56 48 20 4E   45 3E 0A 20  |E|
       0072  20 20 20 20 20 20 E29482     20 20 20 20 20   20 20 20 20  |␣|
       0088  20 20 20 20 37 20     2B     30 20 20 20 20 CE94 20 32 68  |␣|
       0104  20 33 33 6D 20 20     20 EFAA8F 20 2D 35 20 C2B0 43 20 20  |␣|
    """

    def __init__(self, char_per_line: int = 16):
        super().__init__(char_per_line)
        self._output_filters = [NonPrintsStringVisualizer(keep_newlines=False)]

    def _process(self, part: IT):
        self._state.add_row(self._make_row(part))

    def _make_row(self, part: str) -> t.List[str]:
        return [
            " ",
            self._format_offset(),
            " " + self._get_vert_sep_char(),
            "0x",
            " ",
            *self._format_main(part),
            self._get_vert_sep_char(),
            self._format_output_text(part),
        ]

    def _format_main(self, part: str) -> t.Iterable[str]:
        for s in part:
            yield "".join(f"{b:02X}" for b in s.encode())
            yield " "
        yield from [""] * 2 * (self._char_per_line - len(part))

    @classmethod
    def estimate_line_len(cls, char_per_line: int, inp_size: int) -> int:
        space_len = 3
        sep_len = 6
        prefix_len = 2
        offset_len = len(str(inp_size))
        main_len = (8 + 1) * char_per_line
        return space_len + sep_len + prefix_len + offset_len + main_len


class StringUcpTracer(AbstractStringTracer):
    """
    str as Unicode codepoints

    .. todo::
        venv/lib/python3.8/site-packages/pygments/lexers/hexdump.py

    .. code-block:: bash
       :caption: Example output

        56 |U+ 45 4d 20 43 50 55   20   4f 56 48 20 4e  45 3e 0a 20 | EM␣CPU␣OVH␣NE>↵␣
        72 |U+ 20 20 20 20 20 20 2502   20 20 20 20 20  20 20 20 20 | ␣␣␣␣␣␣│␣␣␣␣␣␣␣␣␣
        88 |U+ 20 20 20 20 37 20   2b   30 20 20 20 20 394 20 32 68 | ␣␣␣␣7␣+0␣␣␣␣Δ␣2h
       104 |U+ 20 33 33 6d 20 20   20 fa8f 20 2d 35 20  b0 43 20 20 | ␣33m␣␣␣摒␣-5␣°C␣␣
    """

    def __init__(self, char_per_line: int = 16):
        super().__init__(char_per_line)
        self._output_filters = [NonPrintsStringVisualizer(keep_newlines=False)]

    def _process(self, part: IT):
        self._state.add_row(self._make_row(part))

    def _make_row(self, part: str) -> t.List[str]:
        return [
            " ",
            self._format_offset(),
            " " + self._get_vert_sep_char(),
            "U+",
            " ",
            *self._format_main(part),
            self._get_vert_sep_char(),
            self._format_output_text(part),
        ]

    def _format_main(self, part: str) -> t.Iterable[str]:
        for s in part:
            yield from [f"{ord(s):>02x}", " "]
        yield from [""] * 2 * (self._char_per_line - len(part))

    @classmethod
    def estimate_line_len(cls, char_per_line: int, inp_size: int) -> int:
        space_len = 3
        sep_len = 2
        prefix_len = 2
        offset_len = len(str(inp_size))
        main_len = (5 + 1) * char_per_line
        return space_len + sep_len + prefix_len + offset_len + main_len


@dataclass
class TracerExtra:
    label: str


@dataclass
class _TracerState:
    inp_size: int = field(init=False, default=None)
    lineno: int = field(init=False, default=None)
    offset: int = field(init=False, default=None)

    inp_size_len: int = field(init=False, default=None)
    offset_len: int = field(init=False, default=None)

    rows: t.List[t.List[str]] = field(init=False, default=None)
    cols_max_len: t.List[int] | None = field(init=False, default=None)

    def reset(self, inp: IT):
        self.inp_size = len(inp)
        self.lineno = 0
        self.offset = 0

        self.inp_size_len = 0
        self.offset_len = 0

        self.rows = []
        self.cols_max_len = None

    def add_row(self, row: t.List):
        if self.cols_max_len is None:
            self.cols_max_len = [0] * len(row)
        for col_idx, col_val in enumerate(row):
            self.cols_max_len[col_idx] = max(self.cols_max_len[col_idx], len(col_val))
        self.rows.append(row)


# -----------------------------------------------------------------------------
# Filters[Replacers]


class StringReplacer(IFilter[str, str]):
    """
    .
    """

    def __init__(self, pattern: PT[str], repl: RT[str]):
        if isinstance(pattern, str):
            self._pattern: t.Pattern[str] = re.compile(pattern)
        else:
            self._pattern: t.Pattern[str] = pattern
        self._repl = repl

    def apply(self, inp: str, extra: t.Any = None) -> str:
        return self._replace(inp)

    def _replace(self, inp: str) -> str:
        return self._pattern.sub(self._repl, inp)


class EscSeqStringReplacer(StringReplacer):
    """ """

    def __init__(self, repl: RT[str] = ""):
        super().__init__(ESCAPE_SEQ_REGEX, repl)


class SgrStringReplacer(StringReplacer):
    """
    Find all `SGR <SequenceSGR>` seqs (e.g., ``ESC [1;4m``) and replace with given string. More
    specific version of :class:`CsiReplacer`.

    :param repl:
        Replacement, can contain regexp groups (see :meth:`apply_filters()`).
    """

    def __init__(self, repl: RT[str] = ""):
        super().__init__(SGR_SEQ_REGEX, repl)


class CsiStringReplacer(StringReplacer):
    """
    Find all `CSI <SequenceCSI>` seqs (i.e., starting with ``ESC [``) and replace
    with given string. Less specific version of :class:`SgrReplacer`, as CSI
    consists of SGR and many other sequence subtypes.

    :param repl:
        Replacement, can contain regexp groups (see :meth:`apply_filters()`).
    """

    def __init__(self, repl: RT[str] = ""):
        super().__init__(CSI_SEQ_REGEX, repl)


# -----------------------------------------------------------------------------
# Filters[Mappers]


class OmniMapper(IFilter[IT, IT]):
    """
    Input type: *str*, *bytes*. Abstract mapper. Replaces every character found in
    map keys to corresponding map value. Map should be a dictionary of this type:
    ``dict[int, str|bytes|None]``; moreover, length of *str*/*bytes* must be strictly 1
    character (ASCII codepage). If there is a necessity to map Unicode characters,
    `StringMapper` should be used instead.

    >>> OmniMapper({0x20: '.'}).apply(b'abc def ghi')
    b'abc.def.ghi'

    For mass mapping it is better to subclass `OmniMapper` and override two methods --
    `_get_default_keys` and `_get_default_replacer`. In this case you don't have to
    manually compose a replacement map with every character you want to replace.

    :param override: a dictionary with mappings: keys must be *ints*, values must be
                     either a single-char *strs* or *bytes*, or None.
    :see: `NonPrintsOmniVisualizer`
    """

    def __init__(self, override: MT = None):
        self._make_maps(override)

    def _get_default_keys(self) -> t.List[int]:
        """
        Helper method for avoiding character map manual composing in the mapper subclass.

        :return: List of int codes that should be replaced by default (i.e., without
                 taking ``override`` argument into account, or when it is not present).
        """
        return []

    def _get_default_replacer(self) -> IT:
        """
        Helper method for avoiding character map manual composing in the mapper subclass.

        :return: Default replacement character for int codes that are not present in
                 ``override`` keys list, or when there is no overriding at all.
        """
        raise NotImplementedError

    def _make_maps(self, override: MT | None):
        self._maps = {
            str: str.maketrans(self._make_premap(str, override)),
            bytes: bytes.maketrans(*self._make_bytemaps(override)),
        }

    def _make_premap(self, inp_type: t.Type[IT], override: MT | None) -> t.Dict[int, IT]:
        default_map = dict()
        default_replacer = None
        for i in self._get_default_keys():
            if default_replacer is None:
                default_replacer = self._transcode(
                    self._get_default_replacer(), inp_type
                )
            default_map.setdefault(i, default_replacer)

        if override is None:
            return default_map
        if not isinstance(override, dict):
            raise ArgTypeError(type(override), "override", self.__init__)

        if not all(isinstance(k, int) and 0 <= k <= 255 for k in override.keys()):
            raise TypeError("Mapper keys should be ints such as: 0 <= key <= 255")
        if not all(isinstance(v, (str, bytes)) or v is None for v in override.values()):
            raise TypeError(
                "Each map value should be either a single char in 'str' or 'bytes' form, or None"
            )
        for i, v in override.items():
            default_map.update({i: self._transcode(v, inp_type)})
        return default_map

    def _make_bytemaps(self, override: MT | None) -> t.Tuple[bytes, bytes]:
        premap = self._make_premap(bytes, override)
        srcmap = b"".join(int.to_bytes(k, 1, "big") for k in premap.keys())
        for v in premap.values():
            if len(v) != 1:
                raise ValueError(
                    "All OmniMapper replacement values should be one byte long (i.e. be "
                    "an ASCII char). To utilize non-ASCII characters use StringMapper."
                )
        destmap = b"".join(premap.values())
        return srcmap, destmap

    def apply(self, inp: IT, extra: t.Any = None) -> IT:
        return inp.translate(self._maps[type(inp)])

    def _transcode(self, inp: IT, target: t.Type[RT]) -> RT:
        if isinstance(inp, target):
            return inp
        return inp.encode() if isinstance(inp, str) else inp.decode()


class StringMapper(OmniMapper[str]):
    """
    a
    """

    def _make_maps(self, override: MT | None):
        self._maps = {str: str.maketrans(self._make_premap(str, override))}

    def apply(self, inp: str, extra: t.Any = None) -> str:
        if isinstance(inp, bytes):
            raise TypeError("String mappers allow 'str' as input only")
        return super().apply(inp, extra)


class NonPrintsOmniVisualizer(OmniMapper):
    """
    Input type: *str*, *bytes*. Replace every whitespace character with ``.``.
    """

    def _get_default_keys(self) -> t.List[int]:
        return WHITESPACE_CHARS + CONTROL_CHARS

    def _get_default_replacer(self) -> IT:
        return b"."


class NonPrintsStringVisualizer(StringMapper):
    """
    Input type: *str*. Replace every whitespace character with "·", except
    newlines. Newlines are kept and get prepneded with same char by default,
    but this behaviour can be disabled with ``keep_newlines`` = *False*.

    >>> NonPrintsStringVisualizer().apply('A  B  C')
    'A␣␣B␣␣C'
    >>> apply_filters('1. D'+os.linesep+'2. L ', NonPrintsStringVisualizer(keep_newlines=False))
    '1.␣D↵2.␣L␣'

    :param keep_newlines: When *True*, transform newline characters into "↵\\\\n", or
                          into just "↵" otherwise.
    """

    def __init__(self, keep_newlines: bool = True):
        override = {
            0x09: "⇥",
            0x0A: "↵" + ("\n" if keep_newlines else ""),
            0x0B: "⤓",
            0x0C: "↡",
            0x0D: "⇤",
            0x20: "␣",
        }
        super().__init__(override)

    def _get_default_keys(self) -> t.List[int]:
        return WHITESPACE_CHARS + CONTROL_CHARS

    def _get_default_replacer(self) -> str:
        return "·"


class OmniSanitizer(OmniMapper):
    """
    Input type: *str*, *bytes*. Replace every control character and every non-ASCII
    character (0x80-0xFF) with ".", or with specified char. Note that the replacement
    should be a single ASCII character, because ``Omni-`` filters are designed to work
    with *str* inputs and *bytes* inputs on equal terms.

    :param repl: Value to replace control/non-ascii characters with. Should be strictly 1
                 character long.
    """

    def __init__(self, repl: IT = b"."):
        self._override_replacer = repl
        super().__init__()

    def _get_default_keys(self) -> t.List[int]:
        return CONTROL_CHARS + NON_ASCII_CHARS

    def _get_default_replacer(self) -> IT:
        return self._override_replacer


# -----------------------------------------------------------------------------


def apply_filters(string: IT, *args: AT) -> OT:
    """
    Method for applying dynamic filter list to a target string/bytes.
    Example (will replace all ``ESC`` control characters to ``E`` and
    thus make SGR params visible):

    >>> from pytermor import SeqIndex
    >>> apply_filters(f'{SeqIndex.RED}test{SeqIndex.COLOR_OFF}', SgrStringReplacer(r'E\2\3\4'))
    'E[31mtestE[39m'

    Note that type of ``s`` argument must be same as ``StringFilter`` parameterized
    type, i.e. :class:`ReplaceNonAsciiBytes` is ``StringFilter`` type, so
    you can apply it only to bytes-type strings.

    :param string: String to filter.
    :param args:   `OmniFilter` instance(s) or ``OmniFilter`` type(s).
    :return:       Filtered ``s``.
    """

    def instantiate(f):
        return f() if isinstance(f, type) else f

    return reduce(lambda s, f: instantiate(f)(s), args, string)


def dump(data: t.Any, label: str = None, max_len_shift: int = None) -> str | None:
    """
    .. todo ::
        - format selection
        - special handling of one-line input
        - squash repeating lines
    """
    printer_t = StringUcpTracer
    printer = _dump_printers_cache.get(printer_t, None)

    if not printer and max_len_shift is not None:
        max_len = get_terminal_width() + max_len_shift
        for chars_per_line in [8, 12, 16, 20, 24, 32, 40, 48, 64]:
            est_len = printer_t.estimate_line_len(chars_per_line, len(data))
            if est_len < max_len:
                printer = printer_t(chars_per_line)
                continue
            break

    if not printer:
        printer = printer_t()
    _dump_printers_cache[printer_t] = printer

    return printer.apply(data, TracerExtra(label))



# -----------------------------------------------------------------------------
#  pytermor [ANSI formatted terminal output toolset]
#  (c) 2022-2023. A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
"""
Output formatters. Default global renderer type is `SgrRenderer`.
"""
from __future__ import annotations

import enum
import os
import sys
import typing as t
from abc import abstractmethod, ABCMeta
from functools import reduce
from hashlib import md5

from .ansi import SequenceSGR, NOOP_SEQ, SeqIndex, enclose
from .color import IColor, Color16, Color256, ColorRGB, NOOP_COLOR
from .common import logger, FT
from .style import Style, NOOP_STYLE, Styles, make_style
from .utilmisc import get_qname
from .utilstr import SgrStringReplacer


T = t.TypeVar("T", bound="IRenderer")


def _digest(fingerprint: str) -> int:
    return int.from_bytes(md5(fingerprint.encode()).digest(), "big")


class RendererManager:
    """
    Class for global rendering mode setup.

    Selecting the renderer can be accomplished in several ways:

        a. By using general-purpose functions `text.render()` and `text.echo()` --
           both have an argument ``renderer`` (preferrable; introduced in `pytermor 2.x`).
        b. Method `RendererManager.set_default()` sets the default renderer globally.
           After that calling `text.render()` will automatically invoke a said renderer
           and apply the required formatting (that is, if ``renderer`` argument is
           left empty).
        c. Alternatively, you can use renderer's instance method
           `render() <IRenderer.render>` directly and avoid messing up with the manager,
           but that's not recommended and possibly will be deprecated in future versions).

    Generally speaking, if you need to invoke a custom renderer just once, it's
    convenient to use the first method for this matter, and use the second one
    in all the other cases.

    On the contrary, if there is a necessity to use more than one renderer
    alternatingly, it's better to avoid using the global one at all, and just
    instantiate and invoke both renderers independently.

    .. rubric :: TL;DR

    To unconditionally print formatted message to standard output, do something like
    this:

        >>> from pytermor import render, RendererManager, Styles
        >>> RendererManager.set_default_format_always()
        >>> render('Warning: AAAA', Styles.WARNING)
        '\x1b[33mWarning: AAAA\x1b[39m'

    """

    _default: IRenderer = None

    @classmethod
    def set_default(cls, renderer: IRenderer | t.Type[IRenderer] = None):
        # noinspection PyUnresolvedReferences
        """
        Select a global renderer.

            >>> RendererManager.set_default(SgrRendererDebugger(OutputMode.XTERM_16))
            >>> render('text', Style(fg='red'))
            '(ǝ[31m)text(ǝ[39m)'

        :param renderer:
            Default renderer to use globally. Calling this method without arguments
            will result in library default renderer `SgrRenderer` being set as default.

            All the methods with the ``renderer`` argument (e.g., `text.render()`)
            will use the global default one if said argument is omitted or set to *None*.

            You can specify either the renderer class, in which case manager will
            instantiate it with the default parameters, or provide already instantiated
            and set up renderer, which will be registred as global.
        """
        if isinstance(renderer, type):
            renderer = renderer()
        cls._default = renderer or SgrRenderer()

    @classmethod
    def get_default(cls) -> IRenderer:
        """
        Get global renderer instance (`SgrRenderer`, or the one provided earlier with
        `set_default()`).
        """
        return cls._default

    @classmethod
    def set_default_format_always(cls):
        """
        Shortcut for forcing all control sequences to be present in the
        output of a global renderer.

        Note that it applies only to the renderer that is set up as default at
        the moment of calling this method, i.e., all previously created instances,
        as well as the ones that will be created afterwards, are unaffected.
        """
        cls.set_default(SgrRenderer(OutputMode.TRUE_COLOR))

    @classmethod
    def set_default_format_never(cls):
        """
        Shortcut for disabling all output formatting of a global renderer.
        """
        cls.set_default(SgrRenderer(OutputMode.NO_ANSI))


class IRenderer(metaclass=ABCMeta):
    """Renderer interface."""

    def __hash__(self) -> int:
        """
        Method returning a unique number reflecting current renderer's state. Used for
        rendered strings caching. Two renderers of the same class and with the same
        settings should have equal hashes, so that cached strings could be reused.
        When the internal state of the renderer changes, this number should change as
        well, in order to invalidate the caches.
        """

    @property
    @abstractmethod
    def is_caching_allowed(self) -> bool:
        """
        Class-level property.

        :return: *True* if caching of renderer's results makes any sense and *False*
                 otherwise.
        """

    @property
    @abstractmethod
    def is_format_allowed(self) -> bool:
        """
        :return: *True* if renderer is set up to use the formatting and will do
                 it on invocation, and *False* otherwise.
        """

    @abstractmethod
    def render(self, string: str, fmt: FT = None) -> str:
        """
        Apply colors and attributes described in ``fmt`` argument to
        ``string`` and return the result. Output format depends on renderer's
        class, which defines the implementation.

        :param string: String to format.
        :param fmt:    Style or color to apply. If ``fmt`` is a `IColor` instance,
                       it is assumed to be a foreground color. See `FT`.
        :return: String with formatting applied, or without it, depending on
                 renderer settings.
        """

    @abstractmethod
    def clone(self: T, *args: t.Any, **kwargs: t.Any) -> T:
        """
        Make a copy of the renderer with the same setup.

        :rtype: self
        """

    def __repr__(self):
        return self.__class__.__qualname__ + "[]"


class OutputMode(enum.Enum):
    """
    Determines what types of SGR sequences are allowed to use in the output.
    """

    NO_ANSI = "no_ansi"
    """
    The renderer discards all color and format information completely.
    """
    XTERM_16 = "xterm_16"
    """
    16-colors mode. Enforces the renderer to approximate all color types
    to `Color16` and render them as basic mode selection SGR sequences
    (``ESC [31m``, ``ESC [42m`` etc). See `Color.approximate()` for approximation
    algorithm details.
    """
    XTERM_256 = "xterm_256"
    """
    256-colors mode. Allows the renderer to use either `Color16` or `Color256` 
    (but RGB will be approximated to 256-color pallette).
    """
    TRUE_COLOR = "true_color"
    """
    RGB color mode. Does not apply restrictions to color rendering.
    """
    AUTO = "auto"
    """
    Lets the renderer select the most suitable mode by itself.
    See `SgrRenderer` constructor documentation for the details. 
    """


class SgrRenderer(IRenderer):
    """
    Default renderer invoked by `Text.render()`. Transforms `IColor` instances
    defined in ``style`` into ANSI control sequence bytes and merges them with
    input string. Type of resulting `SequenceSGR` depends on type of `IColor`
    instances in ``style`` argument and current output mode of the renderer.

    1. `ColorRGB` can be rendered as True Color sequence, 256-color sequence
       or 16-color sequence depending on specified `OutputMode`.
    2. `Color256` can be rendered as 256-color sequence or 16-color
       sequence.
    3. `Color16` will be rendered as 16-color sequence.
    4. Nothing of the above will happen and all formatting will be discarded
       completely if output device is not a terminal emulator or if the developer
       explicitly set up the renderer to do so (`OutputMode.NO_ANSI`).

    Renderer approximates RGB colors to closest **indexed** colors if terminal doesn't
    support RGB output. In case terminal doesn't support even 256 colors, it
    falls back to 16-color palette and picks closest samples again the same way.
    See `OutputMode` documentation for exact mappings.

    >>> SgrRenderer(OutputMode.XTERM_256).render('text', Styles.WARNING_LABEL)
    '\x1b[1;33mtext\x1b[22;39m'
    >>> SgrRenderer(OutputMode.NO_ANSI).render('text', Styles.WARNING_LABEL)
    'text'

    :param output_mode:
        SGR output mode to use. Valid values are listed in `OutputMode` enum.

        With `OutputMode.AUTO` the renderer will first check if the output
        device is a terminal emulator, and use `OutputMode.NO_ANSI` when it
        is not. Otherwise, the renderer will read ``TERM`` environment
        variable and follow these rules:

            - `OutputMode.NO_ANSI` if ``TERM`` is set to ``xterm``.
            - `OutputMode.XTERM_16` if ``TERM`` is set to ``xterm-color``.
            - `OutputMode.XTERM_256` in all other cases.

        Special case is when ``TERM`` equals to ``xterm-256color`` **and**
        ``COLORTERM`` is either ``truecolor`` or  ``24bit``, then
        `OutputMode.TRUE_COLOR` will be used.
    """

    _COLOR_UPPER_BOUNDS: t.Dict[OutputMode, t.Type[IColor]] = {
        OutputMode.XTERM_16: Color16,
        OutputMode.XTERM_256: Color256,
        OutputMode.TRUE_COLOR: ColorRGB,
    }

    def __init__(self, output_mode: OutputMode = OutputMode.AUTO, io: t.IO = sys.stdout):
        self._output_mode: OutputMode = self._determine_output_mode(output_mode, io)
        self._color_upper_bound: t.Type[IColor] | None = self._COLOR_UPPER_BOUNDS.get(
            self._output_mode, None
        )

        logger.debug(
            f"Instantiated {self.__class__.__qualname__}"
            f"[{self._output_mode.name} <- {output_mode.name}, "
            f"upper bound {get_qname(self._color_upper_bound)}]"
        )

    def __hash__(self) -> int:
        # although this renderer is immutable, its state can be set up differently
        # on initialization. ``_color_upper_bound`` is a derived variable from
        # ``_output_mode`` with one-to-one mapping, thus it can be omitted.
        return _digest(self.__class__.__qualname__ + "." + self._output_mode.value)

    @property
    def is_caching_allowed(self) -> bool:
        return True

    @property
    def is_format_allowed(self) -> bool:
        return self._output_mode is not OutputMode.NO_ANSI

    def render(self, string: str, fmt: FT = None) -> str:
        style = make_style(fmt)
        opening_seq = (
            self._render_attributes(style, squash=True)
            + self._render_color(style.fg, False)
            + self._render_color(style.bg, True)
        )

        # in case there are line breaks -- split text to lines and apply
        # SGRs for each line separately. it increases the chances that style
        # will be correctly displayed regardless of implementation details of
        # user's pager, multiplexer, terminal emulator etc.
        rendered_text = ""
        for line in string.splitlines(keepends=True):
            rendered_text += enclose(opening_seq, line)
        return rendered_text

    def clone(self) -> SgrRenderer:
        return SgrRenderer(self._output_mode)

    def _determine_output_mode(self, arg_value: OutputMode, io: t.IO) -> OutputMode:
        if arg_value is not OutputMode.AUTO:
            return arg_value

        isatty = io.isatty()
        term = os.environ.get("TERM", None)
        colorterm = os.environ.get("COLORTERM", None)

        logger.debug(f"{io.name} is a terminal: {isatty}")
        logger.debug(f"Environment: TERM='{term}'")
        logger.debug(f"Environment: COLORTERM='{colorterm}'")

        if not isatty:
            return OutputMode.NO_ANSI
        if term == "xterm":
            return OutputMode.NO_ANSI
        if term == "xterm-color":
            return OutputMode.XTERM_16
        if colorterm in ("truecolor", "24bit"):
            return OutputMode.TRUE_COLOR
        return OutputMode.XTERM_256

    def _render_attributes(
        self, style: Style, squash: bool
    ) -> t.List[SequenceSGR] | SequenceSGR:
        if not self.is_format_allowed:
            return NOOP_SEQ if squash else [NOOP_SEQ]

        result = []
        if style.blink:
            result += [SeqIndex.BLINK_SLOW]
        if style.bold:
            result += [SeqIndex.BOLD]
        if style.crosslined:
            result += [SeqIndex.CROSSLINED]
        if style.dim:
            result += [SeqIndex.DIM]
        if style.double_underlined:
            result += [SeqIndex.DOUBLE_UNDERLINED]
        if style.inversed:
            result += [SeqIndex.INVERSED]
        if style.italic:
            result += [SeqIndex.ITALIC]
        if style.overlined:
            result += [SeqIndex.OVERLINED]
        if style.underlined:
            result += [SeqIndex.UNDERLINED]

        if squash:
            return reduce(lambda p, c: p + c, result, NOOP_SEQ)
        return result

    def _render_color(self, color: IColor, bg: bool) -> SequenceSGR:
        if not self.is_format_allowed or color == NOOP_COLOR:
            return NOOP_SEQ
        return color.to_sgr(bg, self._color_upper_bound)


class TmuxRenderer(IRenderer):
    """
    Translates `Styles <Style>` attributes into
    `tmux-compatible <https://man7.org/linux/man-pages/man1/tmux.1.html#STYLES>`_
    markup. `tmux <https://github.com/tmux/tmux>`_ is a commonly used terminal
    multiplexer.

    >>> TmuxRenderer().render('text',  Style(fg='blue', bold=True))
    '#[fg=blue bold]text#[fg=default nobold]'

    """

    STYLE_ATTR_TO_TMUX_MAP = {
        "fg": "fg",
        "bg": "bg",
        "blink": "blink",
        "bold": "bold",
        "crosslined": "strikethrough",
        "dim": "dim",
        "double_underlined": "double-underscore",
        "inversed": "reverse",
        "italic": "italics",
        "overlined": "overline",
        "underlined": "underscore",
    }

    def __hash__(self) -> int:  # stateless
        return _digest(self.__class__.__qualname__)

    @property
    def is_caching_allowed(self) -> bool:
        return True

    @property
    def is_format_allowed(self) -> bool:
        """
        :returns: Always *True*, because tmux markup can be used without regard
                  to the type of output device and its capabilities -- all the
                  dirty work will be done by the multiplexer itself.
        """
        return True

    def render(self, string: str, fmt: FT = None) -> str:
        style = make_style(fmt)
        command_open, command_close = self._render_attributes(style)
        rendered_text = ""
        for line in string.splitlines(keepends=True):
            rendered_text += command_open + line + command_close
        return rendered_text

    def clone(self) -> TmuxRenderer:
        return TmuxRenderer()

    def _render_attributes(self, style: Style) -> t.Tuple[str, ...]:
        cmd_open: t.List[t.Tuple[str, str]] = []
        cmd_close: t.List[t.Tuple[str, str]] = []

        for attr_name, tmux_name in self.STYLE_ATTR_TO_TMUX_MAP.items():
            attr_val = getattr(style, attr_name)
            if attr_val is None:
                continue
            if isinstance(attr_val, IColor):
                if attr_val == NOOP_COLOR:
                    continue
                cmd_open.append((tmux_name + "=", attr_val.to_tmux(attr_name == "bg")))
                cmd_close.append((tmux_name + "=", "default"))
            elif isinstance(attr_val, bool):
                if not attr_val:
                    continue
                cmd_open.append((tmux_name, ""))
                cmd_close.append(("no" + tmux_name, ""))
            else:
                raise TypeError(
                    f"Unexpected attribute type: {type(attr_val)} for '{attr_name}'"
                )
        return self._encode_tmux_command(cmd_open), self._encode_tmux_command(cmd_close)

    def _encode_tmux_command(self, kv: t.List[t.Tuple[str, str]]) -> str:
        if len(kv) == 0:
            return ""
        return "#[" + (" ".join(f"{k}{v}" for k, v in kv)) + "]"


class NoOpRenderer(IRenderer):
    """
    Special renderer type that does nothing with the input string and just
    returns it as is. Often used as a default argument value (along with similar
    "NoOps" like `NOOP_STYLE`, `NOOP_COLOR` etc.)

    >>> NoOpRenderer().render('text', Style(fg='green', bold=True))
    'text'

    """

    def __bool__(self) -> bool:
        return False

    def __hash__(self) -> int:  # stateless
        return _digest(self.__class__.__qualname__)

    @property
    def is_caching_allowed(self) -> bool:
        return False

    @property
    def is_format_allowed(self) -> bool:
        """
        :returns: Nothing to apply |rarr| nothing to allow, thus the returned value
                  is always *False*.
        """
        return False

    def render(self, string: str, fmt: FT = None) -> str:
        """
        Return the `string` argument untouched, don't mind the `fmt`.

        :param string: String to :strike:`format` ignore.
        :param fmt:    Style or color to :strike:`appl`  discard.
        """
        return string

    def clone(self) -> NoOpRenderer:
        return NoOpRenderer()


class HtmlRenderer(IRenderer):
    """
    Translate `Styles <Style>` attributes into a rudimentary HTML markup.
    All the formatting is inlined into ``style`` attribute of the ``<span>``
    elements. Can be optimized by extracting the common styles as CSS classes
    and referencing them by DOM elements instead.

    >>> HtmlRenderer().render('text', Style(fg='red', bold=True))
    '<span style="color: #800000; font-weight: 700">text</span>'

    """

    DEFAULT_ATTRS = [
        "color",
        "background-color",
        "font-weight",
        "font-style",
        "text-decoration",
        "border",
        "filter",
    ]

    def __hash__(self) -> int:  # stateless
        return _digest(self.__class__.__qualname__)

    @property
    def is_caching_allowed(self) -> bool:
        return True

    @property
    def is_format_allowed(self) -> bool:
        """
        :returns: Always *True*, because the capabilities of the terminal have
                  nothing to do with HTML markup meant for web-browsers.
        """
        return True

    def render(self, string: str, fmt: FT = None) -> str:
        style = make_style(fmt)
        opening_tag, closing_tag = self._render_attributes(style)
        return f"{opening_tag}{string}{closing_tag}"  # @TODO  # attribues

    def clone(self) -> HtmlRenderer:
        return HtmlRenderer()

    def _render_attributes(self, style: Style = NOOP_STYLE) -> t.Tuple[str, str]:
        if style == NOOP_STYLE:
            return "", ""

        span_styles: t.Dict[str, t.Set[str]] = dict()
        for attr in self._get_default_attrs():
            span_styles[attr] = set()

        if style.fg != NOOP_COLOR:
            span_styles["color"].add(style.fg.format_value("#"))
        if style.bg != NOOP_COLOR:
            span_styles["background-color"].add(style.bg.format_value("#"))

        if style.blink:  # modern browsers doesn't support it without shit piled up
            span_styles["border"].update(("1px", "dotted"))
        if style.bold:
            span_styles["font-weight"].add("700")
        if style.crosslined:
            span_styles["text-decoration"].add("line-through")
        if style.dim:
            span_styles["filter"].update(("saturate(0.5)", "brightness(0.75)"))
        if style.double_underlined:
            span_styles["text-decoration"].update(("underline", "double"))
        if style.inversed:
            span_styles["color"], span_styles["background-color"] = (
                span_styles["background-color"],
                span_styles["color"],
            )
        if style.italic:
            span_styles["font-style"].add("italic")
        if style.overlined:
            span_styles["text-decoration"].add("overline")
        if style.underlined:
            span_styles["text-decoration"].add("underline")

        span_class_str = (
            "" if style.class_name is None else f' class="{style.class_name}"'
        )
        span_style_str = "; ".join(
            f"{k}: {' '.join(v)}" for k, v in span_styles.items() if len(v) > 0
        )
        return f'<span{span_class_str} style="{span_style_str}">', "</span>"

    def _get_default_attrs(self) -> t.List[str]:
        return self.DEFAULT_ATTRS


class SgrRendererDebugger(SgrRenderer):
    """
    Subclass of regular `SgrRenderer` with two differences -- instead of rendering the
    proper ANSI escape sequences it renders them with ``ESC`` character replaced by "ǝ",
    and encloses the whole sequence into '()' for visual separation.

    Can be used for debugging of assembled sequences, because such a transformation
    reliably converts a control sequence into a harmless piece of bytes completely ignored
    by the terminals.

    >>> SgrRendererDebugger(OutputMode.XTERM_16).render('text', Style(fg='red', bold=True))
    '(ǝ[1;31m)text(ǝ[22;39m)'

    """

    def __init__(self, output_mode: OutputMode = OutputMode.AUTO):
        super().__init__(output_mode)
        self._format_override: bool | None = None

    def __hash__(self) -> int:
        # build the hash from instance's state as well as ancestor's state -- that way
        # it will reflect the changes in either of configurations. actually, sometimes
        # the hashes will be different, but the results would have been the same;
        # e.g., `SgrRendererDebugger` with ``_format_override`` set to *False* and
        # `SgrRendererDebugger` without the override and with `NO_ANSI` output mode
        # has different hashes, but produce exactly the same outputs. however,
        # this can be disregarded, as it is not worth the efforts to implement an
        # advanced logic and correct state computation when it comes to debug renderer.
        return _digest(
            ".".join(
                [
                    self.__class__.__qualname__,
                    str(self._format_override),
                    str(super().__hash__()),
                ]
            )
        )

    @property
    def is_caching_allowed(self) -> bool:
        return True

    @property
    def is_format_allowed(self) -> bool:
        if self._format_override is not None:
            return self._format_override
        return super().is_format_allowed

    def render(self, string: str, fmt: FT = None) -> str:
        return SgrStringReplacer(r"(ǝ\2\3\4)").apply(super().render(string, fmt))

    def clone(self) -> SgrRendererDebugger:
        cloned = SgrRendererDebugger(self._output_mode)
        cloned._format_override = self._format_override
        return cloned

    def set_format_always(self):
        """
        Force all control sequences to be present in the output.
        """
        self._format_override = True

    def set_format_auto(self):
        """
        Reset the force formatting flag and let the renderer decide by itself (see
        `SgrRenderer` docs for the details).
        """
        self._format_override = None

    def set_format_never(self):
        """
        Force disabling of all output formatting.
        """
        self._format_override = False


RendererManager.set_default()



# -----------------------------------------------------------------------------
#  pytermor [ANSI formatted terminal output toolset]
#  (c) 2022-2023. A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
"""
Module contains definitions for low-level ANSI escape sequences building.
Can be used for creating a variety of sequences including:

    - :abbr:`SGR (Select Graphic Rendition)` sequences (text and background
      coloring, other text formatting and effects);
    - :abbr:`CSI (Control Sequence Introducer)` sequences (cursor management,
      selective screen cleraing);
    - :abbr:`OSC (Operating System Command)` sequences (varoius system commands).

.. important ::
    blah-blah-blah low-level @TODO

"""
from __future__ import annotations

from abc import abstractmethod, ABCMeta
from copy import copy
import enum
import typing as t

from .common import ConflictError


class ISequence(t.Sized, metaclass=ABCMeta):
    """
    Abstract ancestor of all escape sequences.
    """

    _ESC_CHARACTER: str = "\x1b"
    _SEPARATOR: str = ";"
    _INTRODUCER: str
    _TERMINATOR: str

    def __init__(self, *params: int | str):
        """
        :param  \*params:  Sequence internal parameters, existnce and
                           valid amount depends on sequence type.
        """
        self._params: t.List[int | str] = []

        for param in params:
            if isinstance(param, IntCode):
                self._params.append(
                    param.value
                )  # to avoid casting it later in assemble()
                continue
            self._params.append(param)

    def assemble(self) -> str:
        """
        Build up actual byte sequence and return
        as an ASCII-encoded string.
        """
        return (
            self._ESC_CHARACTER
            + self._INTRODUCER
            + self._SEPARATOR.join([str(param) for param in self.params])
            + self._TERMINATOR
        )

    @property
    def params(self) -> t.List[int | str]:
        """
        Return internal params as array.
        """
        return self._params

    @classmethod
    @abstractmethod
    def _short_class_name(cls):
        raise NotImplementedError

    def __str__(self) -> str:
        return self.assemble()

    def __len__(self) -> int:
        return 0

    def __eq__(self, other: ISequence):
        if type(self) != type(other):
            return False
        return self._params == other._params

    def __repr__(self) -> str:
        params = ",".join([str(p) for p in self._params])
        return f"<{self._short_class_name()}[{params}]>"


class ISequenceFe(ISequence, metaclass=ABCMeta):
    """
    Wide range of sequence types that includes `CSI <SequenceCSI>`,
    `OSC <SequenceOSC>` and more.

    All subtypes of this sequence start with ``ESC`` plus ASCII byte
    from ``0x40`` to ``0x5F`` (``@``, ``[``, ``\\``, ``]``, ``_``, ``^`` and
    capital letters ``A``-``Z``).
    """


class SequenceST(ISequenceFe):
    """
    String Terminator sequence (ST). Terminates strings in other control
    sequences. Encoded as ``ESC \\`` (``0x1B`` ``0x5C``).
    """

    _INTRODUCER = "\\"

    def assemble(self) -> str:
        return self._ESC_CHARACTER + self._INTRODUCER

    @classmethod
    def _short_class_name(cls) -> str:
        return "ST"


class SequenceOSC(ISequenceFe):
    """
    :abbr:`OSC (Operating System Command)`-type sequence. Starts a control
    string for the operating system to use. Encoded as ``ESC ]``, plus params
    separated by ``;``, and terminated with `SequenceST`.
    """

    _INTRODUCER = "]"
    _TERMINATOR = SequenceST().assemble()

    @classmethod
    def _short_class_name(cls):
        return "OSC"


class SequenceCSI(ISequenceFe):
    """
    Class representing :abbr:`CSI (Control Sequence Introducer)`-type ANSI
    escape sequence. All subtypes of this sequence start with ``ESC [``.

    Sequences of this type are used to control text formatting,
    change cursor position, erase screen and more.

    >>> make_erase_in_line().assemble()
    '\x1b[0K'

    """

    _INTRODUCER = "["

    def __init__(self, terminator: str, short_name: str, *params: int):
        """

        :param terminator:
        :param short_name:
        :param params:
        """
        self._terminator = terminator
        self._short_name = short_name
        super().__init__(*params)

    def regexp(self) -> str:
        return f"\\x1b\\[[0-9;]*{self._terminator}"

    def assemble(self) -> str:
        return (
            self._ESC_CHARACTER
            + self._INTRODUCER
            + self._SEPARATOR.join([str(param) for param in self.params])
            + self._terminator
        )

    def _short_class_name(self):
        result = "CSI"
        if self._short_name:
            result += ":" + self._short_name
        return result


class SequenceSGR(SequenceCSI):
    """
    Class representing :abbr:`SGR (Select Graphic Rendition)`-type escape sequence
    with varying amount of parameters. SGR sequences allow to change the color
    of text or/and terminal background (in 3 different color spaces) as well
    as set decorate text with italic style, underlining, overlining, cross-lining,
    making it bold or blinking etc.

    When cast to *str*, as all other sequences, invokes `assemble()` method
    and transforms into encoded control sequence string. It is possible to add
    of one SGR sequence to another, resulting in a new one with merged params
    (see examples).

    .. note ::
        `SequenceSGR` with zero params was specifically implemented to
        translate into empty string and not into ``ESC [m``, which would have
        made sense, but also would be entangling, as this sequence is the equivalent
        of ``ESC [0m`` -- hard reset sequence. The empty-string-sequence is
        predefined at module level as `NOOP_SEQ`.

    .. note ::
        The module doesn't distinguish "single-instruction" sequences from several
        ones merged together, e.g. ``Style(fg='red', bold=True)`` produces only one
        opening SequenceSGR instance:

        >>> SequenceSGR(IntCode.BOLD, IntCode.RED).assemble()
        '\x1b[1;31m'

        ...although generally speaking it is two of them (``ESC [1m`` and
        ``ESC [31m``). However, the module can automatically match terminating
        sequences for any form of input SGRs and translate it to specified format.

    >>> SequenceSGR(IntCode.HI_CYAN, 'underlined', 1)
    <SGR[96,4,1]>
    >>> SequenceSGR(31) + SequenceSGR(1) == SequenceSGR(31, 1)
    True

    :param args:  ..  ::

                    Sequence params. Resulting param order is the same as an
                    argument order. Each argument can be specified as:

                     - *str* -- any of `IntCode` names, case-insensitive
                     - *int* -- `IntCode` instance or plain integer
                     - `SequenceSGR` instance (params will be extracted)
    """

    _TERMINATOR = "m"

    def __init__(self, *args: str | int | SequenceSGR):
        result: t.List[int] = []

        for arg in args:
            if isinstance(arg, str):
                result.append(IntCode.resolve(arg).value)
            elif isinstance(arg, int):
                result.append(arg)
            elif isinstance(arg, SequenceSGR):
                result.extend(arg.params)
            else:
                raise TypeError(f"Invalid argument type: {arg!r})")

        result = [max(0, p) for p in result]
        super().__init__(self._TERMINATOR, "SGR", *result)

    def assemble(self) -> str:
        if len(self._params) == 0:  # NOOP
            return ""

        return (
            self._ESC_CHARACTER
            + self._INTRODUCER
            + self._SEPARATOR.join([str(param) for param in self._params])
            + self._TERMINATOR
        )

    @property
    def params(self) -> t.List[int]:
        """
        :return: Sequence params as integers or `IntCode` instances.
        """
        return super().params

    def __hash__(self) -> int:
        return int.from_bytes(self.assemble().encode(), byteorder="big")

    def __add__(self, other: SequenceSGR) -> SequenceSGR:
        self._ensure_sequence(other)
        return SequenceSGR(*self._params, *other._params)

    def __radd__(self, other: SequenceSGR) -> SequenceSGR:
        return other.__add__(self)

    def __iadd__(self, other: SequenceSGR) -> SequenceSGR:
        return self.__add__(other)

    def __eq__(self, other: SequenceSGR):
        if type(self) != type(other):
            return False
        return self._params == other._params

    @staticmethod
    def _ensure_sequence(subject: t.Any):
        if not isinstance(subject, SequenceSGR):
            raise TypeError(f"Expected SequenceSGR, got {type(subject)}")

    @staticmethod
    def validate_extended_color(value: int):
        if value < 0 or value > 255:
            raise ValueError(
                f"Invalid color value: expected range [0-255], got: {value}"
            )

    @classmethod
    def _short_class_name(cls) -> str:
        return "SGR"


# class UnderlinedCurlySequenceSGR(SequenceSGR):
#     """
#     Registered as a separate class, because this is the one and only SGR in the
#     package, which is identified by "4:3" string (in contrast with all the other
#     sequences entirely made of digits and semicolon separators).
#     """
#
#     def __init__(self):
#         """ """
#         super().__init__()
#         self._params = [f"{IntCode.UNDERLINED}:3"]
#
#     @property
#     def params(self) -> t.List[str]:
#         """ """
#         return self._params

class _NoOpSequenceSGR(SequenceSGR):
    def __init__(self):
        super().__init__()

    def __bool__(self) -> bool:
        return False

    def __eq__(self, other: SequenceSGR) -> bool:
        if not isinstance(other, SequenceSGR):
            return False
        return self._params == other._params

    def __repr__(self) -> str:
        return f"<{self._short_class_name()}[NOP]>"


NOOP_SEQ = _NoOpSequenceSGR()
"""
Special sequence in case you *have to* provide one or another SGR, but do 
not want any control sequences to be actually included in the output. 
``NOOP_SEQ.assemble()`` returns empty string, ``NOOP_SEQ.params`` 
returns empty list.

>>> NOOP_SEQ.assemble()
''
>>> NOOP_SEQ.params
[]

"""


class IntCode(enum.IntEnum):
    """
    Complete or almost complete list of reliably working SGR param integer codes.
    Fully interchangeable with plain *int*. Suitable for `SequenceSGR`
    default constructor.

    .. note ::
        `IntCode` predefined constants are omitted from documentation to avoid
        useless repeats and save space, as most of the time "next level" class
        `SeqIndex` is more appropriate, and on top of that, the constant
        names are literally the same for `SeqIndex` and `IntCode`.
    """

    @classmethod
    def resolve(cls, name: str) -> IntCode:
        """

        :param name:
        """
        name_norm = name.upper()
        try:
            instance = cls[name_norm]
        except KeyError as e:
            e.args = (f"Int code '{name_norm}' (<- '{name}') does not exist",)
            raise e
        return instance

    def __repr__(self) -> str:
        return f"<{self.value}|{self.name}>"

    # -- SGR: default attributes and colors -----------------------------------

    RESET = 0  # hard reset code
    BOLD = 1
    DIM = 2
    ITALIC = 3
    UNDERLINED = 4
    # CURLY_UNDERLINED = '4:3'  # TODO
    BLINK_SLOW = 5
    BLINK_FAST = 6
    INVERSED = 7
    HIDDEN = 8
    CROSSLINED = 9
    DOUBLE_UNDERLINED = 21
    OVERLINED = 53
    BOLD_DIM_OFF = 22  # no sequence to disable BOLD or DIM while keeping the other
    ITALIC_OFF = 23
    UNDERLINED_OFF = 24
    BLINK_OFF = 25
    INVERSED_OFF = 27
    HIDDEN_OFF = 28
    CROSSLINED_OFF = 29
    COLOR_OFF = 39
    BG_COLOR_OFF = 49
    OVERLINED_OFF = 55

    BLACK = 30
    RED = 31
    GREEN = 32
    YELLOW = 33
    BLUE = 34
    MAGENTA = 35
    CYAN = 36
    WHITE = 37
    COLOR_EXTENDED = 38

    BG_BLACK = 40
    BG_RED = 41
    BG_GREEN = 42
    BG_YELLOW = 43
    BG_BLUE = 44
    BG_MAGENTA = 45
    BG_CYAN = 46
    BG_WHITE = 47
    BG_COLOR_EXTENDED = 48

    # UNDERLINE_COLOR_EXTENDED = 58  # @TODO
    # UNDERLINE_COLOR_OFF = 59

    GRAY = 90
    HI_RED = 91
    HI_GREEN = 92
    HI_YELLOW = 93
    HI_BLUE = 94
    HI_MAGENTA = 95
    HI_CYAN = 96
    HI_WHITE = 97

    BG_GRAY = 100
    BG_HI_RED = 101
    BG_HI_GREEN = 102
    BG_HI_YELLOW = 103
    BG_HI_BLUE = 104
    BG_HI_MAGENTA = 105
    BG_HI_CYAN = 106
    BG_HI_WHITE = 107

    # RARELY SUPPORTED (thus excluded)
    # 10-20: font selection
    #    26: proportional spacing
    #    50: disable proportional spacing
    #    51: framed
    #    52: encircled
    #    54: neither framed nor encircled
    # 60-65: ideogram attributes
    # 73-75: superscript and subscript

    # -- SGR: extended modifiers ----------------------------------------------

    EXTENDED_MODE_256 = 5
    EXTENDED_MODE_RGB = 2

    # -- Other sequence classes -----------------------------------------------

    HYPERLINK = 8


class SeqIndex:
    """
    Registry of static sequence presets.
    """

    # -- SGR ------------------------------------------------------------------

    RESET = SequenceSGR(IntCode.RESET)
    """
    Hard reset sequence.
    """

    BOLD = SequenceSGR(IntCode.BOLD)
    """ Bold or increased intensity. """

    DIM = SequenceSGR(IntCode.DIM)
    """ Faint, decreased intensity. """

    ITALIC = SequenceSGR(IntCode.ITALIC)
    """ Italic *(not widely supported)*. """

    UNDERLINED = SequenceSGR(IntCode.UNDERLINED)
    """ Underline. """

    BLINK_SLOW = SequenceSGR(IntCode.BLINK_SLOW)
    """ Set blinking to < 150 cpm. """

    BLINK_FAST = SequenceSGR(IntCode.BLINK_FAST)
    """ Set blinking to 150+ cpm *(not widely supported)*. """

    INVERSED = SequenceSGR(IntCode.INVERSED)
    """ Swap foreground and background colors. """

    HIDDEN = SequenceSGR(IntCode.HIDDEN)
    """ Conceal characters *(not widely supported)*. """

    CROSSLINED = SequenceSGR(IntCode.CROSSLINED)
    """ Strikethrough. """

    DOUBLE_UNDERLINED = SequenceSGR(IntCode.DOUBLE_UNDERLINED)
    """ Double-underline. *On several terminals disables* `BOLD` *instead*. """

    OVERLINED = SequenceSGR(IntCode.OVERLINED)
    """ Overline *(not widely supported)*. """

    BOLD_DIM_OFF = SequenceSGR(IntCode.BOLD_DIM_OFF)
    """
    Disable ``BOLD`` and ``DIM`` attributes.\n
    *Special aspects... It's impossible to reliably disable them on a separate basis.*
    """

    ITALIC_OFF = SequenceSGR(IntCode.ITALIC_OFF)
    """ Disable italic. """

    UNDERLINED_OFF = SequenceSGR(IntCode.UNDERLINED_OFF)
    """ Disable underlining. """

    BLINK_OFF = SequenceSGR(IntCode.BLINK_OFF)
    """ Disable blinking. """

    INVERSED_OFF = SequenceSGR(IntCode.INVERSED_OFF)
    """ Disable inversing. """

    HIDDEN_OFF = SequenceSGR(IntCode.HIDDEN_OFF)
    """ Disable conecaling. """

    CROSSLINED_OFF = SequenceSGR(IntCode.CROSSLINED_OFF)
    """ Disable strikethrough. """

    OVERLINED_OFF = SequenceSGR(IntCode.OVERLINED_OFF)
    """ Disable overlining. """

    # text colors

    BLACK = SequenceSGR(IntCode.BLACK)
    """ Set text color to 0x000000. """

    RED = SequenceSGR(IntCode.RED)
    """ Set text color to 0x800000. """

    GREEN = SequenceSGR(IntCode.GREEN)
    """ Set text color to 0x008000. """

    YELLOW = SequenceSGR(IntCode.YELLOW)
    """ Set text color to 0x808000. """

    BLUE = SequenceSGR(IntCode.BLUE)
    """ Set text color to 0x000080. """

    MAGENTA = SequenceSGR(IntCode.MAGENTA)
    """ Set text color to 0x800080. """

    CYAN = SequenceSGR(IntCode.CYAN)
    """ Set text color to 0x008080. """

    WHITE = SequenceSGR(IntCode.WHITE)
    """ Set text color to 0xc0c0c0. """

    COLOR_OFF = SequenceSGR(IntCode.COLOR_OFF)
    """ Reset foreground color. """

    # background colors

    BG_BLACK = SequenceSGR(IntCode.BG_BLACK)
    """ Set background color to 0x000000. """

    BG_RED = SequenceSGR(IntCode.BG_RED)
    """ Set background color to 0x800000. """

    BG_GREEN = SequenceSGR(IntCode.BG_GREEN)
    """ Set background color to 0x008000. """

    BG_YELLOW = SequenceSGR(IntCode.BG_YELLOW)
    """ Set background color to 0x808000. """

    BG_BLUE = SequenceSGR(IntCode.BG_BLUE)
    """ Set background color to 0x000080. """

    BG_MAGENTA = SequenceSGR(IntCode.BG_MAGENTA)
    """ Set background color to 0x800080. """

    BG_CYAN = SequenceSGR(IntCode.BG_CYAN)
    """ Set background color to 0x008080. """

    BG_WHITE = SequenceSGR(IntCode.BG_WHITE)
    """ Set background color to 0xc0c0c0. """

    BG_COLOR_OFF = SequenceSGR(IntCode.BG_COLOR_OFF)
    """ Reset background color. """

    # high intensity text colors

    GRAY = SequenceSGR(IntCode.GRAY)
    """ Set text color to 0x808080. """
    HI_RED = SequenceSGR(IntCode.HI_RED)
    """ Set text color to 0xff0000. """
    HI_GREEN = SequenceSGR(IntCode.HI_GREEN)
    """ Set text color to 0x00ff00. """
    HI_YELLOW = SequenceSGR(IntCode.HI_YELLOW)
    """ Set text color to 0xffff00. """
    HI_BLUE = SequenceSGR(IntCode.HI_BLUE)
    """ Set text color to 0x0000ff. """
    HI_MAGENTA = SequenceSGR(IntCode.HI_MAGENTA)
    """ Set text color to 0xff00ff. """
    HI_CYAN = SequenceSGR(IntCode.HI_CYAN)
    """ Set text color to 0x00ffff. """
    HI_WHITE = SequenceSGR(IntCode.HI_WHITE)
    """ Set text color to 0xffffff. """

    # high intensity background colors

    BG_GRAY = SequenceSGR(IntCode.BG_GRAY)
    """ Set background color to 0x808080. """
    BG_HI_RED = SequenceSGR(IntCode.BG_HI_RED)
    """ Set background color to 0xff0000. """
    BG_HI_GREEN = SequenceSGR(IntCode.BG_HI_GREEN)
    """ Set background color to 0x00ff00. """
    BG_HI_YELLOW = SequenceSGR(IntCode.BG_HI_YELLOW)
    """ Set background color to 0xffff00. """
    BG_HI_BLUE = SequenceSGR(IntCode.BG_HI_BLUE)
    """ Set background color to 0x0000ff. """
    BG_HI_MAGENTA = SequenceSGR(IntCode.BG_HI_MAGENTA)
    """ Set background color to 0xff00ff. """
    BG_HI_CYAN = SequenceSGR(IntCode.BG_HI_CYAN)
    """ Set background color to 0x00ffff. """
    BG_HI_WHITE = SequenceSGR(IntCode.BG_HI_WHITE)
    """ Set background color to 0xffffff. """

    # -- OSC ------------------------------------------------------------------

    HYPERLINK = SequenceOSC(IntCode.HYPERLINK)
    """
    Create a hyperlink in the text *(supported by limited amount of terminals)*.
    Note that for a working hyperlink you'll need two sequences, not just one.
     
    .. seealso ::
        `make_hyperlink_part()` and `assemble_hyperlink()`.
    """


COLORS = list(range(30, 39))
BG_COLORS = list(range(40, 49))
HI_COLORS = list(range(90, 98))
BG_HI_COLORS = list(range(100, 108))
ALL_COLORS = COLORS + BG_COLORS + HI_COLORS + BG_HI_COLORS


class _SgrPairityRegistry:
    """
    Internal class providing methods for mapping SGRs to a
    complement (closing) SGRs, also referred to as "resetters".
    """

    _code_to_resetter_map: t.Dict[int | t.Tuple[int, ...], SequenceSGR] = dict()
    _complex_code_def: t.Dict[int | t.Tuple[int, ...], int] = dict()
    _complex_code_max_len: int = 0

    def __init__(self):
        _regulars = [
            (IntCode.BOLD, IntCode.BOLD_DIM_OFF),
            (IntCode.DIM, IntCode.BOLD_DIM_OFF),
            (IntCode.ITALIC, IntCode.ITALIC_OFF),
            (IntCode.UNDERLINED, IntCode.UNDERLINED_OFF),
            (IntCode.DOUBLE_UNDERLINED, IntCode.UNDERLINED_OFF),
            (IntCode.BLINK_SLOW, IntCode.BLINK_OFF),
            (IntCode.BLINK_FAST, IntCode.BLINK_OFF),
            (IntCode.INVERSED, IntCode.INVERSED_OFF),
            (IntCode.HIDDEN, IntCode.HIDDEN_OFF),
            (IntCode.CROSSLINED, IntCode.CROSSLINED_OFF),
            (IntCode.OVERLINED, IntCode.OVERLINED_OFF),
        ]

        for c in _regulars:
            self._bind_regular(*c)

        for c in [*COLORS, *HI_COLORS]:
            self._bind_regular(c, IntCode.COLOR_OFF)

        for c in [*BG_COLORS, *BG_HI_COLORS]:
            self._bind_regular(c, IntCode.BG_COLOR_OFF)

        self._bind_complex((IntCode.COLOR_EXTENDED, 5), 1, IntCode.COLOR_OFF)
        self._bind_complex((IntCode.COLOR_EXTENDED, 2), 3, IntCode.COLOR_OFF)
        self._bind_complex((IntCode.BG_COLOR_EXTENDED, 5), 1, IntCode.BG_COLOR_OFF)
        self._bind_complex((IntCode.BG_COLOR_EXTENDED, 2), 3, IntCode.BG_COLOR_OFF)

    def _bind_regular(self, starter_code: int | t.Tuple[int, ...], resetter_code: int):
        if starter_code in self._code_to_resetter_map:
            raise ConflictError(f"SGR {starter_code} already has a registered resetter")

        self._code_to_resetter_map[starter_code] = SequenceSGR(resetter_code)

    def _bind_complex(
        self, starter_codes: t.Tuple[int, ...], param_len: int, resetter_code: int
    ):
        self._bind_regular(starter_codes, resetter_code)

        if starter_codes in self._complex_code_def:
            raise ConflictError(f"SGR {starter_codes} already has a registered resetter")

        self._complex_code_def[starter_codes] = param_len
        self._complex_code_max_len = max(
            self._complex_code_max_len, len(starter_codes) + param_len
        )

    def get_closing_seq(self, opening_seq: SequenceSGR) -> SequenceSGR:
        closing_seq_params: t.List[int] = []
        opening_params = copy(opening_seq.params)

        while len(opening_params):
            key_params: int | t.Tuple[int, ...] | None = None

            for complex_len in range(
                1, min(len(opening_params), self._complex_code_max_len + 1)
            ):
                opening_complex_suggestion = tuple(opening_params[:complex_len])

                if opening_complex_suggestion in self._complex_code_def:
                    key_params = opening_complex_suggestion
                    complex_total_len = (
                        complex_len + self._complex_code_def[opening_complex_suggestion]
                    )
                    opening_params = opening_params[complex_total_len:]
                    break

            if key_params is None:
                key_params = opening_params.pop(0)
            if key_params not in self._code_to_resetter_map:
                continue

            closing_seq_params.extend(self._code_to_resetter_map[key_params].params)

        return SequenceSGR(*closing_seq_params)


_sgr_pairity_registry = _SgrPairityRegistry()


def get_closing_seq(opening_seq: SequenceSGR) -> SequenceSGR:
    """

    :param opening_seq:
    :return:
    """
    return _sgr_pairity_registry.get_closing_seq(opening_seq)


def enclose(opening_seq: SequenceSGR, string: str) -> str:
    """

    :param opening_seq:
    :param string:
    :return:
    """
    return f"{opening_seq}{string}{get_closing_seq(opening_seq)}"


def make_set_cursor_x_abs(x: int = 1) -> SequenceCSI:
    """
    Create :abbr:`CHA (Cursor Horizontal Absolute)` sequence that sets
    cursor horizontal position, or column, to ``x``.

    :param x:  New cursor horizontal position.
    :example:  ``ESC [1G``
    """
    if x <= 0:
        raise ValueError(f"Invalid x value: {x}, expected x > 0")
    return SequenceCSI("G", "CHA", x)


def make_erase_in_line(mode: int = 0) -> SequenceCSI:
    """
    Create :abbr:`EL (Erase in Line)` sequence that erases a part of the line
    or the entire line. Cursor position does not change.

    :param mode:  .. ::

                  Sequence operating mode.

                     - If set to 0, clear from cursor to the end of the line.
                     - If set to 1, clear from cursor to beginning of the line.
                     - If set to 2, clear the entire line.

    :example:     ``ESC [0K``
    """
    if not (0 <= mode <= 2):
        raise ValueError(f"Invalid mode: {mode}, expected [0;2]")
    return SequenceCSI("K", "EL", mode)


def make_query_cursor_position() -> SequenceCSI:
    """
    Create :abbr:`QCP (Query Cursor Position)` sequence that requests an output
    device to respond with a structure containing current cursor coordinates
    (`RCP <decompose_request_cursor_position()>`).

    .. warning ::

        Sending this sequence to the terminal may **block** infinitely. Consider
        using a thread or set a timeout for the main thread using a signal.

    :example:   ``ESC [6n``
    """

    return SequenceCSI("n", "QCP", 6)


def decompose_request_cursor_position(string: str) -> t.Tuple[int, int]|None:
    """
    Parse :abbr:`RCP (Report Cursor Position)` sequence that generally comes from
    a terminal as a response to `QCP <make_query_cursor_position>` sequence and
    contains a cursor's current row and column.

    .. note ::

        As the library in general provides sequence assembling methods, but not
        the disassembling ones, there is no dedicated class for RCP sequences yet.

    >>> decompose_request_cursor_position('\x1b[18;2R')
    (18, 2)

    :param string:  Terminal response with a sequence.
    :return:        Current row and column if the expected sequence exists
                    in ``string``, *None* otherwise.
    """
    if not all(c in string for c in "\x1b[;R"):
        return None  # input does not contain a valid RCP sequence

    start_idx = string.rfind("\x1b[")
    seq = string[start_idx + 2: string.rfind("R") - start_idx]
    y, x = seq.split(";", 2)

    try:
        return int(y), int(x)
    except ValueError:
        return None  # unexpected results


def make_color_256(code: int, bg: bool = False) -> SequenceSGR:
    """
    Wrapper for creation of `SequenceSGR` that sets foreground
    (or background) to one of 256-color palette value.

    :param code:  Index of the color in the palette, 0 -- 255.
    :param bg:    Set to *True* to change the background color
                  (default is foreground).
    :example:     ``ESC [38;5;141m``
    """

    SequenceSGR.validate_extended_color(code)
    key_code = IntCode.BG_COLOR_EXTENDED if bg else IntCode.COLOR_EXTENDED
    return SequenceSGR(key_code, IntCode.EXTENDED_MODE_256, code)


def make_color_rgb(r: int, g: int, b: int, bg: bool = False) -> SequenceSGR:
    """
    Wrapper for creation of `SequenceSGR` operating in True Color mode (16M).
    Valid values for ``r``, ``g`` and ``b`` are in range of [0; 255]. This range
    linearly translates into [0x00; 0xFF] for each channel. The result
    value is composed as "0xRRGGBB". For example, sequence with color of
    0xFF3300 can be created with::

        make_color_rgb(255, 51, 0)

    :param r:  Red channel value, 0 -- 255.
    :param g:  Blue channel value, 0 -- 255.
    :param b:  Green channel value, 0 -- 255.
    :param bg: Set to *True* to change the background color (default is foreground).
    :example:  ``ESC [38;2;255;51;0m``
    """

    [SequenceSGR.validate_extended_color(color) for color in [r, g, b]]
    key_code = IntCode.BG_COLOR_EXTENDED if bg else IntCode.COLOR_EXTENDED
    return SequenceSGR(key_code, IntCode.EXTENDED_MODE_RGB, r, g, b)


def make_hyperlink_part(url: str = None) -> SequenceOSC:
    """

    :param url:
    :example: ``ESC ]8;;http://localhost ESC \\``
    """
    return SequenceOSC(IntCode.HYPERLINK, "", (url or ""))


def assemble_hyperlink(url: str, label: str = None) -> str:
    """

    :param url:
    :param label:
    :example:  ``ESC ]8;;http://localhost ESC \\Text ESC ]8;; ESC \\``
    """
    return f"{make_hyperlink_part(url)}{label or url}{make_hyperlink_part()}"



# -----------------------------------------------------------------------------
#  pytermor [ANSI formatted terminal output toolset]
#  (c) 2022-2023. A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
"""
utilnum
"""
from __future__ import annotations

import re
import typing as t
from dataclasses import dataclass
from math import floor, log10, trunc, log, isclose

from .common import RT, logger, Align
from .cval import CVAL as cv
from .style import NOOP_STYLE, Style, Styles
from .text import Text, Fragment, IRenderable
from .utilstr import pad

_OVERFLOW_CHAR = "!"


def format_thousand_sep(val: int | float, separator: str = " ") -> str:
    """
    Returns input ``val`` with integer part split into groups of three digits,
    joined then with ``separator`` string.

    >>> format_thousand_sep(260341)
    '260 341'
    >>> format_thousand_sep(-9123123123.55, ',')
    '-9,123,123,123.55'

    :param val:
    :param separator:
    """
    return f"{val:_}".replace("_", separator)


# -----------------------------------------------------------------------------


def format_auto_float(val: float, req_len: int, allow_exp_form: bool = True) -> str:
    """
    Dynamically adjust decimal digit amount and format to fill up the output string
    with as many significant digits as possible, and keep the output length strictly
    equal to `req_len`  at the same time.

    For values impossible to fit into a string of required length and when rounding
    doesn't help (e.g. 12 500 000 and 5 chars) algorithm switches to scientific notation,
    and the result looks like '1.2e7', unless this feature is explicitly disabled
    with ``allow_exp_form`` = *False*; in that case:

        1) if absolute value is less than 1, zeros will be displayed ('0.0000');
        2) if value is a big number (like :math:`10^9`), *ValueError* will be
           raised instead.

    >>> format_auto_float(0.012345678, 5)
    '0.012'
    >>> format_auto_float(0.123456789, 5)
    '0.123'
    >>> format_auto_float(1.234567891, 5)
    '1.235'
    >>> format_auto_float(12.34567891, 5)
    '12.35'
    >>> format_auto_float(123.4567891, 5)
    '123.5'
    >>> format_auto_float(1234.567891, 5)
    ' 1235'
    >>> format_auto_float(12345.67891, 5)
    '12346'

    :param val:             Value to format.
    :param req_len:         Required output string length.
    :param allow_exp_form:  Allow scientific notation usage when no other way
                            of fitting the value into a string of required length.
    :raises ValueError:     If value is too big to fit into ``req_len`` digits and
                            ``allow_exp_form`` is set to False.
    """
    if req_len < -1:
        raise ValueError(f"Required length should be >= 0 (got {req_len})")

    sign = ""
    if val < 0:
        sign = "-"
        req_len -= 1

    if req_len == 0:
        return _OVERFLOW_CHAR * (len(sign))

    abs_value = abs(val)
    if abs_value < 1 and req_len == 1:
        # '0' is better than '-'
        return f"{sign}0"

    if val == 0.0:
        return f"{sign}{0:{req_len}.0f}"

    oom = floor(log10(abs_value))  # order of magnitude
    exp_threshold_left = -2
    exp_threshold_right = req_len - 1

    # oom threshold depends on req
    # length on the right, and is
    # fixed to -2 on the left:

    #   req |    3     4     5     6 |
    #  thld | -2,2  -2,3  -2,4  -2,5 |

    # it determines oom values to
    # enable scientific notation for:

    #  value exp req thld cnd result |
    # ------ --- --- ---- --- ------ |
    # 0.0001  -4   4 -2,3   t   1e-4 |
    #  0.001  -3   4 -2,3   f   1e-3 | - less than threshold
    #   0.01  -2   4 -2,3   f   0.01 | ---- N -
    #    0.1  -1   4 -2,3   f   0.10 |      O M
    #      1   0   4 -2,3   f   1.00 |      R O
    #     10   1   4 -2,3   f   10.0 |      M D
    #    100   2   4 -2,3   f  100.0 |      A E
    #   1000   3   4 -2,3   f   1000 | ---- L -
    #  10000   4   4 -2,3   t    1e4 | - greater than threshold

    if not allow_exp_form:
        # if exponent mode is disabled, the formatter will try as best
        # as he can to display at least something meaningful; although
        # this can work for the values with the order of magnitude < 0
        # (and result in something like '0.001'), the method will fail
        # on the large input values.
        exp_threshold_left = None

    require_exp_form = (
        exp_threshold_left is not None and oom < exp_threshold_left
    ) or oom > exp_threshold_right

    if require_exp_form:
        if not allow_exp_form:  # oh well...
            msg = f"Failed to fit {val:.2f} into {req_len} chars without exp notation"
            raise ValueError(msg)

        exponent_len = len(str(oom)) + 1  # 'e'
        if req_len < exponent_len:
            # there is no place even for exponent
            return _OVERFLOW_CHAR * (len(sign) + req_len)

        significand = abs_value / pow(10, oom)
        max_significand_len = req_len - exponent_len
        try:
            # max_significand_len can be 0, in that case significand_str will be empty; that
            # means we cannot fit it the significand, but still can display approximate number power
            # using the 'eN'/'-eN' notation
            significand_str = format_auto_float(
                significand, max_significand_len, allow_exp_form=False
            )

        except ValueError:
            return f"{sign}e{oom}".rjust(req_len)

        return f"{sign}{significand_str}e{oom}"

    integer_len = max(1, oom + 1)
    if integer_len == req_len:
        # special case when rounding
        # can change the result length
        integer_str = f"{abs_value:{req_len}.0f}"

        if len(integer_str) > integer_len:
            # e.g. req_len = 1, abs_value = 9.9
            #      => should be displayed as 9, not 10
            integer_str = f"{trunc(abs_value):{req_len}d}"

        return f"{sign}{integer_str}"

    decimals_with_point_len = req_len - integer_len
    decimals_len = decimals_with_point_len - 1

    # dot without decimals makes no sense, but
    # python's standard library handles
    # this by itself: f'{12.3:.0f}' => '12'
    dot_str = f".{decimals_len!s}"

    return f"{sign}{abs_value:{req_len}{dot_str}f}"


# -----------------------------------------------------------------------------


class NumHighlighter:
    # fmt: off
    PREFIX_UNIT_REGEX = re.compile(
        r"""
        (?: ^ | (?<!\x1b\[)(?<=[]()\s\[-])\b )
        (?P<intp>\d+)
        (?P<frac>[.,]\d+)?
        (?P<sep>\s?)
        (?P<prefix>[kmgtpuμµn%])?
        (?P<unit>
            i?b[/ip]?t?s?|
            v|a|s(?:econd|ec|)s?|
            m(?:inute|in|onth|on|o|)s?|
            h(?:our|r|)s?|
            d(?:ay|)s?|
            w(?:eek|k|)s?|
            y(?:ear|r|)s?|
            hz
        )?
        (?= []\s\b()[] | $ )
        """,
        flags=re.IGNORECASE | re.VERBOSE,
    )
                                                                  # |_PW_|_G.MULT___G.DIV______TIME______
    STYLE_DEFAULT = NOOP_STYLE                                    # |    | misc.               second
    STYLE_NUL = Style(STYLE_DEFAULT, dim=True)                    # |  0 | zero
    STYLE_PRC = Style(STYLE_DEFAULT, fg=cv.MAGENTA, bold=True)    # |  2 |          percent
    STYLE_KIL = Style(STYLE_DEFAULT, fg=cv.BLUE, bold=True)       # |  3 | Kilo-    milli-     minute
    STYLE_MEG = Style(STYLE_DEFAULT, fg=cv.CYAN, bold=True)       # |  6 | Mega-    micro-     hour
    STYLE_GIG = Style(STYLE_DEFAULT, fg=cv.GREEN, bold=True)      # |  9 | Giga-    nano-      day
    STYLE_TER = Style(STYLE_DEFAULT, fg=cv.YELLOW, bold=True)     # | 12 | Tera-    pico-      week
    STYLE_MON = Style(STYLE_DEFAULT, fg=cv.HI_YELLOW, bold=True)  # |    |                     month
    STYLE_PET = Style(STYLE_DEFAULT, fg=cv.RED, bold=True)        # | 15 | Peta-               year

    PREFIX_MAP = {
        '%': STYLE_PRC,
        'K': STYLE_KIL, 'k': STYLE_KIL, 'm': STYLE_KIL,
        'M': STYLE_MEG, 'μ': STYLE_MEG, 'µ': STYLE_MEG,
        'G': STYLE_GIG, 'g': STYLE_GIG, 'n': STYLE_GIG,
        'T': STYLE_TER, 'p': STYLE_TER,
        'P': STYLE_PET,
    }
    TIME_UNIT_MAP = {
        '%': STYLE_PRC,
        's': STYLE_PRC, 'sec': STYLE_PRC, 'second': STYLE_PRC,
        'm': STYLE_KIL, 'min': STYLE_KIL, 'minute': STYLE_KIL,
        'h': STYLE_MEG,  'hr': STYLE_MEG,   'hour': STYLE_MEG,
        'd': STYLE_GIG, 'day': STYLE_GIG,
        'w': STYLE_TER,  'wk': STYLE_TER, 'week': STYLE_TER,
        'M': STYLE_MON,  'mo': STYLE_MON,  'mon': STYLE_MON, 'month': STYLE_MON,
        'y': STYLE_PET,  'yr': STYLE_PET, 'year': STYLE_PET,
    }
    # fmt: on

    @classmethod
    def get_prefix_style(cls, prefix: str) -> Style:
        return cls.PREFIX_MAP.get(prefix, cls.STYLE_DEFAULT)

    @classmethod
    def get_time_unit_style(cls, time_unit: str) -> Style:
        if len(time_unit) > 1:
            time_unit = time_unit.removesuffix("s")
        return cls.TIME_UNIT_MAP.get(time_unit, cls.STYLE_DEFAULT)

    @classmethod
    def format(cls, string: str) -> Text:
        cursor = 0
        result = Text()
        for m in cls.PREFIX_UNIT_REGEX.finditer(string):
            result += string[cursor : m.start()]
            result.append(*cls.colorize(**m.groupdict("")))
            cursor = m.end()
        result += string[cursor:]
        return result

    @classmethod
    def colorize(
        cls, intp: str, frac: str, sep: str, prefix: str, unit: str
    ) -> t.List[Fragment]:
        unit_norm = unit.rstrip("s").strip()
        int_st = cls.STYLE_DEFAULT
        if prefix:
            int_st = cls.PREFIX_MAP.get(prefix, cls.STYLE_DEFAULT)
        elif unit_norm:
            int_st = cls.TIME_UNIT_MAP.get(unit_norm, cls.STYLE_DEFAULT)

        digits = intp + frac[1:]
        if digits.count("0") == len(digits):
            int_st = cls.STYLE_NUL

        frac_st = Style(int_st, dim=True)
        unit_st = Style(int_st, dim=True, bold=False)
        return [
            Fragment(intp, int_st),
            Fragment(frac, frac_st),
            Fragment(sep),
            Fragment(prefix + unit, unit_st),
        ]


# -----------------------------------------------------------------------------


class StaticBaseFormatter:
    """
    Format ``value`` using settings passed to constructor. The purpose of this class
    is to fit into specified string length as much significant digits as it's
    theoretically possible by using multipliers and unit prefixes. Designed
    for metric systems with bases 1000 or 1024.

    You can create your own formatters if you need fine tuning of the
    output and customization. If that's not the case, there are facade
    methods :meth:`format_si()` and :meth:`format_si_binary()`,
    which will invoke predefined formatters and doesn't require setting up.
    """

    # fmt: off
    PREFIXES_SI_DEC = [
        'q', 'r', 'y', 'z', 'a', 'f', 'p', 'n', 'μ', 'm',
        None,
        'k', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y', 'R', 'Q',
    ]
    """
    Prefix preset used by `format_si()`. Covers values from :math:`10^{-30}` to 
    :math:`10^{32}`. 
    """

    PREFIXES_SI_BIN = [None, 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi', 'Yi', 'Ri', 'Qi']
    """
    Prefix preset used by `format_si_binary()`. Covers values from :math:`0` to 
    :math:`10^{32}`. 
    """
    # fmt: on

    _attribute_defaults = dict(
        _max_value_len=4,
        _color=False,
        _allow_negative=True,
        _allow_fractional=True,
        _discrete_input=False,
        _unit="",
        _unit_separator=" ",
        _mcoef=1000.0,
        _pad=False,
        _legacy_rounding=False,
        _prefixes=PREFIXES_SI_DEC,
        _prefix_refpoint_shift=0,
        _value_mapping=dict(),
    )

    def __init__(
        self,
        fallback: StaticBaseFormatter = None,
        *,
        max_value_len: int = None,
        color: bool = None,
        allow_negative: bool = None,
        allow_fractional: bool = None,
        discrete_input: bool = None,
        unit: str = None,
        unit_separator: str = None,
        mcoef: float = None,
        pad: bool = None,
        legacy_rounding: bool = None,
        prefixes: t.List[str | None] = None,
        prefix_refpoint_shift: int = None,
        value_mapping: t.Dict[float, RT] | t.Callable[[float], RT] = None,  # @TODO
    ):
        """
        .. note ::

            All arguments except ``fallback`` are *kwonly*-type arguments.

        :param fallback: Take missing (i.e., *None*) attribute values from this instance.
        :param int max_value_len:
            [default: 4] Target string length. Must be at least **3**, because it's a
            minimum requirement for formatting values from 0 to 999.
            Next number to 999 is 1000, which will be formatted as "1k".

            Setting ``allow_negative`` to *True* increases lower bound to **4** because
            the values now can be less than 0, and minus sign also occupies one char in
            the output.

            Setting ``mcoef`` to anything other than 1000.0 also increases the minimum
            by 1, to **5**. The reason is that non-decimal coefficients like 1024 require
            additional char to render as switching to the next prefix happens later:
            "999 b", "1000 b", "1001 b", ..."1023 b", "1 Kb".

        :param bool  color: [default: *False*]
        :param bool  allow_negative:
            [default: *True*] Allow negative numbers handling, or (if set to *False*)
            ignore the sign and round all of them to 0.0. This option effectively
            increases lower limit of ``max_value_len`` by 1.
        :param bool  allow_fractional:
            [default: *True*] Allows the usage of fractional values in the output. If
            set to *False*, the results will be rounded.
        :param bool  discrete_input:
            [default: *False*] If set to *True*, truncate the fractional part off the
            input and do not use floating-point format for *base output*, i.e., without
            prefix and multiplying coefficient. Useful when the values are originally
            discrete (e.g., bytes). Note that the same effect could be achieved by
            setting ``allow_fractional`` to *False*, except that it will influence
            prefixed output as well ("1.08 kB" -> "1kB").
        :param str   unit:
            [default: empty *str*] Unit to apply prefix to (e.g., "m", 'B").
            Can be empty.
        :param str   unit_separator:
            [default: a space] String to place in between the value and the
            (prefixed) unit. Can be empty.
        :param float mcoef:
            [default: 1000.0] Multiplying coefficient applied to the value:

            .. math ::

                V_{out} = V_{in} * b^{(-m/3)},

            where: :math:`V_{in}` is an input value, :math:`V_{out}` is a numeric part
            of the output, :math:`b` is ``mcoef`` (base), and :math:`m` is the order of
            magnitude corresponding to a selected unit prefix. For example, in case
            of default (decimal) formatter and input value equal to 17345989 the selected
            prefix will be "M" with the order of magnitude = 6:

            .. math ::

                V_{out} = 17345989*1000^{(-6/3)} = 17345989*10^{-6} = 17.346 .

        :param bool  pad: [default: *False*]
        :param bool  legacy_rounding: [default: *False*]
        :param list[str|None] prefixes:
            [default: `PREFIXES_SI_DEC`] Prefix list from min power to max. Reference
            point (with zero-power multiplier, or 1.0) is determined by searching
            for *None* in the list provided, therefore it's a requirement for the
            argument to have at least one *None* value. Prefix list for a formatter
            without fractional values support could look like this::

                [None, "k", 'M", "G", "T"]

            Prefix step is fixed to :math:`log_{10} 1000 = 3`, as specified for
            metric prefixes.

        :param int prefix_refpoint_shift:
            [default: 0] Should be set to a non-zero number if input represents
            already prefixed value; e.g. to correctly format a variable,
            which stores the frequency in MHz, set prefix shift to 2;
            the formatter then will render 2333 as "2.33 GHz" instead of
            incorrect "2.33 kHz".
        :param value_mapping: @TODO
        """
        self._max_value_len: int = max_value_len
        self._color: bool = color
        self._allow_negative: bool = allow_negative
        self._allow_fractional: bool = allow_fractional
        self._discrete_input: bool = discrete_input
        self._unit: str = unit
        self._unit_separator: str = unit_separator
        self._mcoef: float = mcoef
        self._pad: bool = pad
        self._legacy_rounding: bool = legacy_rounding
        self._prefixes: t.List[str | None] = prefixes
        self._prefix_refpoint_shift: int = prefix_refpoint_shift
        self._value_mapping: t.Dict[float, RT] | t.Callable[[float], RT] = value_mapping

        for attr_name, default in self._attribute_defaults.items():
            if getattr(self, attr_name) is None:
                if (fallback_attr := getattr(fallback, attr_name, None)) is not None:
                    setattr(self, attr_name, fallback_attr)
                    continue
                setattr(self, attr_name, default)

        if self._max_value_len < self.max_len_lower_bound:
            raise ValueError(
                f"Impossible to display all decimal numbers as "
                f"{self._max_value_len}-char length."
            )

        try:
            self._prefix_refpoint_idx: int = self._prefixes.index(None)
        except ValueError:
            raise ValueError(
                "At least one of the prefixes should be None, as it indicates "
                "the reference point with multiplying coefficient 1.00."
            )
        self._prefix_refpoint_idx += self._prefix_refpoint_shift
        if not (0 <= self._prefix_refpoint_idx < len(self._prefixes)):
            raise ValueError("Shifted prefix reference point is out of bounds")

    def get_max_len(self, unit_ov: str = None) -> int:
        """
        :param unit_ov: Unit override. Set to *None* to use formatter's own unit.
        :return:        Maximum length of the result. Note that constructor argument
                        is `max_value_len`, which is a different parameter.
        """
        result = self._max_value_len
        result += len(self._unit_separator)
        result += len(self._get_unit_effective(unit_ov))
        result += self._get_max_prefix_len()
        return result

    @property
    def max_len_lower_bound(self) -> int:
        result = 3
        if self._allow_negative:
            result += 1
        if not self.is_decimal:
            result += 1
        return result

    @property
    def is_decimal(self) -> bool:
        return self._mcoef == 1000.0

    def format(self, val: float, unit_ov: str = None, color_ov: bool = None) -> RT:
        """
        :param val:       Input value.
        :param unit_ov:   Unit override. Set to *None* to use formatter's own unit.
        :param color_ov:  Color mode override, *bool* to enable/disable
                          colorizing, *None* to use formatters' setting value.
        :return: Formatted value, *Text* if colorizing is on, *str* otherwise.
        """
        origin_val = val
        unit = self._get_unit_effective(unit_ov)
        if not self._allow_negative:
            val = max(0.0, val)
        if self._discrete_input and self._prefix_refpoint_shift == 0:
            val = trunc(val)

        abs_value = abs(val)
        power_base = self._mcoef ** (1 / 3)  # =10 for metric, ~10.079 for binary
        if abs_value == 0.0:
            prefix_shift = 0
        else:
            # order of magnitude:
            oom = floor(round(log(abs_value, power_base), 6))
            # rounding is required because cumulative floating point error can lead to:
            # 1024^9 = 1099511600000.0 -> "1024 GiB" (expected "1.00 TiB"),
            # 1023 -> "1.00 KiB" (expected "1023 B").

            oom_shift = 0 if self._legacy_rounding else -1  # "0.18s" --> "180ms"
            prefix_shift = round((oom + oom_shift) / 3)

        base_output = prefix_shift == 0
        val /= power_base ** (prefix_shift * 3)
        unit_idx = self._prefix_refpoint_idx + prefix_shift
        if 0 <= unit_idx < len(self._prefixes):
            prefix = self._prefixes[unit_idx] or ""
        else:
            prefix = "?" * self._get_max_prefix_len()

        sep = self._unit_separator
        fractional_output = self._allow_fractional and not (
            base_output and self._discrete_input
        )

        # drop excessive digits first, or get excessive zeros for the values near float
        # precision limit for 64-bit systems, e.g. for 10 - e-15 (=9.999999999999998) and
        # max_value_len=3 the result would be "10.0" instead of "10".
        eff_val = round(val, self._max_value_len - 1)

        # @TODO here i started to feel that there should be an easier way to do all the
        #       math; more specifically, to avoid one or two redundant transformations,
        #       but it's impossible to prove or reject it without an investigation. the
        #       reason for this is existence of three rounding operations in a row.
        if fractional_output:
            val_str = format_auto_float(eff_val, self._max_value_len, False)
        else:
            val_str = f"{trunc(eff_val):d}"
        if len(val_str) > self._max_value_len:
            logger.warning(
                "Inconsistent result -- max val length %d exceeded (%d): '%s' <- %f"
                % (self._max_value_len, len(val_str), val_str, origin_val)
            )

        result = self._colorize(color_ov, val_str, sep, prefix, unit)
        if self._pad:
            max_len = self.get_max_len(unit_ov)
            pad_len = max(0, max_len - len(result))
            result = pad(pad_len) + result
            if isinstance(result, IRenderable):
                result.set_width(max_len)
        return result

    def _colorize(
        self, color_ov: bool, val: str, sep: str, prefix: str, unit: str
    ) -> RT:
        unit_full = (prefix + unit).strip()
        if not unit_full or unit_full.isspace():
            sep = ""

        if not self._get_color_effective(color_ov):
            return "".join((val.strip(), sep, unit_full))

        int_part, point, frac_part = val.strip().partition(".")
        result = NumHighlighter.colorize(int_part, point + frac_part, sep, prefix, unit)
        return Text(*result, align=Align.RIGHT)

    def _get_unit_effective(self, unit_ov: str) -> str:
        if unit_ov is not None:
            return unit_ov
        return self._unit

    def _get_color_effective(self, color_ov: bool) -> bool:
        if color_ov is not None:
            return color_ov
        return self._color

    def _get_max_prefix_len(self) -> int:
        return max([len(p) for p in self._prefixes if p is not None])

    def __repr__(self) -> str:
        return self.__class__.__qualname__


formatter_si = StaticBaseFormatter()
"""
Decimal SI formatter, formats ``value`` as a unitless value with SI-prefixes;
a unit can be provided as an argument of `format()` method. Suitable for 
formatting any SI unit with values from :math:`10^{-30}` to :math:`10^{32}`.

:usage:

    .. code-block :: 
        
        # either of: 
        formatter_si.format(<value>, ...)
        format_si(<value>, ...)
    
:max len: 
  
    Total maximum length is ``max_value_len + 2``, which is **6**
    by default (4 from value + 1 from separator and + 1 from prefix).
    If the unit is defined and is a non-empty string, the maximum output 
    length increases by length of that unit.

:see: `format_si()`

"""

formatter_si_binary = StaticBaseFormatter(
    allow_negative=False,
    discrete_input=True,
    unit="B",
    unit_separator=" ",
    mcoef=1024.0,
    prefixes=StaticBaseFormatter.PREFIXES_SI_BIN,
)
"""
Binary SI formatter, formats ``value`` as binary size ("KiB", "MiB") 
with base = 1024. Unit can be customized.

While being similar to `formatter_si`, this formatter 
differs in one aspect.  Given a variable with default value = 995,
formatting its value results in "995 B". After increasing it
by 20 we'll have 1015, but it's still not enough to become
a kilobyte -- so returned value will be "1015 B". Only after one
more increase (at 1024 and more) the value will be in a form
of "1.00 KiB".

:usage:

    .. code-block :: 
        
        # either of: 
        formatter_si_binary.format(<value>, ...)
        format_si_binary(<value>, ...)
    
:max len: 
  
    First things first, the initial ``max_value_len`` must be at least 5 (not 4),
    because it is a minimum requirement for formatting values from 1023 to -1023.
    
    The negative values for this formatter are disabled by default and thus 
    will be rounded to 0, which decreases the ``max_value_len`` minimum value 
    by 1 (to 4).
    
    Total maximum length is ``max_value_len + 4`` = **8** (base + 1 from separator,
    1 from unit and 2 from prefix, assuming all of them have default values 
    defined in `formatter_si_binary`).

:see: `format_si_binary()`

"""


def format_si(val: float, unit: str = None, color: bool = None) -> RT:
    """
    Wrapper for `formatter_si.format()<formatter_si>`.

    >>> format_si(1010, 'm²')
    '1.01 km²'
    >>> format_si(0.223, 'g')
    '223 mg'
    >>> format_si(1213531546, 'W')  # great scott
    '1.21 GW'
    >>> format_si(1.22e28, 'eV')  # the Planck energy
    '12.2 ReV'

    :param val:    Input value (unitless).
    :param unit:   A unit override [default unit is an empty string].
    :param color:  If *True*, the result will be colorized depending
                   on prefix type.
    :return: Formatted value, *Text* if colorizing is on, *str* otherwise.
    """
    return formatter_si.format(val, unit, color)


def format_si_binary(val: float, unit: str = None, color: bool = False) -> RT:
    """
    Wrapper for `formatter_si_binary.format()<formatter_si_binary>`.

    >>> format_si_binary(1010)  # 1010 b < 1 kb
    '1010 B'
    >>> format_si_binary(1080)
    '1.05 KiB'
    >>> format_si_binary(45200)
    '44.1 KiB'
    >>> format_si_binary(1.258 * pow(10, 6), 'b')
    '1.20 Mib'

    :param val:    Input value in bytes.
    :param unit:   A unit override [default unit is "B"].
    :param color:  If *True*, the result will be colorized depending
                   on prefix type.
    :return: Formatted value, *Text* if colorizing is on, *str* otherwise.
    """
    return formatter_si_binary.format(val, unit, color)


# -----------------------------------------------------------------------------


def format_time_delta(val_sec: float, max_len: int = None, color_ov: bool = None) -> RT:
    """
    Format time delta using suitable format (which depends on
    ``max_len`` argument). Key feature of this formatter is
    ability to combine two units and display them simultaneously,
    e.g. return "3h 48min" instead of "228 mins" or "3 hours",

    There are predefined formatters with output length of **3, 4,
    6** and **10** characters. Therefore, you can pass in any value
    from 3 inclusive and it's guarenteed that result's length
    will be less or equal to required length. If `max_len` is
    omitted, longest registred formatter will be used.

    >>> format_time_delta(10, 3)
    '10s'
    >>> format_time_delta(10, 6)
    '10.0s'
    >>> format_time_delta(15350, 4)
    '4 h'
    >>> format_time_delta(15350)
    '4h 15min'

    :param val_sec:   Value to format.
    :param max_len:   Maximum output string length (total).
    :param color_ov:  Color mode override, *bool* to enable/disable
                      colorizing depending on unit type, *None* to use formatters'
                      setting value.
    """
    if max_len is None:
        formatter = tdf_registry.get_longest()
    else:
        formatter = tdf_registry.find_matching(max_len)

    if formatter is None:
        raise ValueError(f"No settings defined for max length = {max_len} (or less)")

    return formatter.format(val_sec, color_ov)


class DynamicBaseFormatter:
    """
    Formatter designed for time intervals. Key feature of this formatter is
    ability to combine two units and display them simultaneously,
    e.g. return "3h 48min" instead of "228 mins" or "3 hours", etc.

    It is possible to create custom formatters if fine tuning of the output and
    customization is necessary; otherwise use a facade method `format_time_delta()`,
    which selects appropriate formatter by specified max length from a preset list.

    Example output::

        "10 secs", "5 mins", "4h 15min", "5d 22h"

    :param units:
    :param color: If *True*, the result will be colorized depending on unit type.
    :param allow_negative:
    :param allow_fractional:
    :param unit_separator:
    :param pad:   Set to *True* to pad the value with spaces on the left side
                  and ensure it's length is equal to `max_len`, or to *False*
                  to allow shorter result strings.
    :param plural_suffix:
    :param overflow_msg:
    """

    def __init__(
        self,
        units: t.List[CustomBaseUnit],
        *,
        color: bool = False,
        allow_negative: bool = False,
        allow_fractional: bool = True,
        unit_separator: str = None,
        pad: bool = False,
        plural_suffix: str = None,
        overflow_msg: str = "OVERFLOW",
    ):
        self._units = units
        self._color = color
        self._allow_negative = allow_negative
        self._allow_fractional = allow_fractional
        self._unit_separator = unit_separator
        self._pad = pad
        self._plural_suffix = plural_suffix
        self._overflow_msg = overflow_msg

        self._fractional_formatter = StaticBaseFormatter(
            max_value_len=4,
            unit="s",
            unit_separator="",
            allow_negative=True,
            allow_fractional=True,
        )
        self._max_len = None
        self._compute_max_len()

    @property
    def max_len(self) -> int:
        """
        This property cannot be set manually, it is
        computed on initialization automatically.

        :return: Maximum possible output string length.
        """
        return self._max_len

    def format(self, val: float, color_ov: bool = None) -> RT:
        """
        Pretty-print difference between two moments in time. If input
        value is too big for the current formatter to handle, return "OVERFLOW"
        string (or a part of it, depending on ``max_len``).

        :param val: Input value.
        :param color_ov:  Color mode override, *bool* to enable/disable
                          colorizing, *None* to use formatters' setting value.
        :return: Formatted time delta, *Text* if colorizing is on, *str* otherwise.
        """
        result = self.format_base(val, color_ov)
        if result is None:
            result = self._overflow_msg[: self.max_len]
            if self._get_color_effective(color_ov):
                result = Text(result, Styles.ERROR_LABEL)

        if self._pad:
            pad_len = max(0, self._max_len - len(result))
            result = pad(pad_len) + result
            if isinstance(result, IRenderable):
                result.set_width(self._max_len)

        return result

    # @TODO naming?
    def format_base(self, val: float, color_ov: bool = None) -> RT | None:
        """
        Pretty-print difference between two moments in time. If input
        value is too big for the current formatter to handle, return *None*.

        :param val:   Input value.
        :param color_ov:  Color mode override, *bool* to enable/disable
                          colorizing, *None* to use formatters' setting value.
        :return: Formatted value as *Text* if colorizing is on; as *str*
                 otherwise. Returns *None* on overflow.
        """
        num = abs(val)
        unit_idx = 0
        result_sub = ""

        negative = self._allow_negative and val < 0
        sign = "-" if negative else ""
        result = None

        if self._allow_fractional and num < self._units[0].in_next:
            if self._max_len is not None:
                result = self._fractional_formatter.format(val, color_ov=color_ov)
                if len(result) > self._max_len:
                    # for example, 500ms doesn't fit in the shortest possible
                    # delta string (which is 3 chars), so "<1s" will be returned
                    result = None

        while result is None and unit_idx < len(self._units):
            unit = self._units[unit_idx]
            if unit.overflow_after and num > unit.overflow_after:
                if not self._max_len:  # max len is being computed now
                    raise RecursionError()
                return None

            unit_name = unit.name
            unit_name_suffixed = unit_name
            if self._plural_suffix and trunc(num) != 1:
                unit_name_suffixed += self._plural_suffix

            unit_short = unit_name[0]
            if unit.custom_short:
                unit_short = unit.custom_short

            next_unit_ratio = unit.in_next
            sep = self._unit_separator or ""

            if abs(num) < 1:
                if negative:
                    result = self._colorize(color_ov, "~", "0", sep, unit_name_suffixed)
                elif isclose(num, 0, abs_tol=1e-03):
                    result = self._colorize(color_ov, "", "0", sep, unit_name_suffixed)
                else:
                    result = self._colorize(color_ov, "<", "1", sep, unit_name)

            elif unit.collapsible_after is not None and num < unit.collapsible_after:
                val = str(floor(num))
                result = self._colorize(color_ov, sign, val, "", unit_short) + result_sub

            elif not next_unit_ratio or num < next_unit_ratio:
                val = str(floor(num))
                result = self._colorize(color_ov, sign, val, sep, unit_name_suffixed)

            else:
                next_num = floor(num / next_unit_ratio)
                prev_val = str(floor(num - (next_num * next_unit_ratio)))
                result_sub = self._colorize(color_ov, sep, prev_val, "", unit_short)
                num = next_num
                unit_idx += 1
                continue

        return result or ""

    def _colorize(self, color_ov: bool, extra: str, val: str, sep: str, unit: str) -> RT:
        if not self._get_color_effective(color_ov):
            return "".join((extra, val, sep, unit))

        return Text(
            Fragment(extra),
            *NumHighlighter.colorize(val, "", sep, "", unit),
            align=Align.RIGHT,
        )

    def _get_color_effective(self, color_ov: bool) -> bool:
        if color_ov is not None:
            return color_ov
        return self._color

    def _compute_max_len(self):
        max_len = 0
        coef = 1.00

        for unit in self._units:
            test_val = unit.in_next
            if not test_val:
                test_val = unit.overflow_after
            if not test_val:
                continue
            test_val_sec = coef * (test_val - 1) * (-1 if self._allow_negative else 1)

            try:
                max_len_unit = self.format_base(test_val_sec, color_ov=False)
            except RecursionError:
                continue

            max_len = max(max_len, len(max_len_unit))
            coef *= unit.in_next or unit.overflow_after

        self._max_len = max_len


@dataclass(frozen=True)
class CustomBaseUnit:
    """
    TU

    .. important ::

        ``in_next`` and `overflow_after` are mutually exclusive, and either of
        them is required.

    :param name: A unit name to display.
    :param in_next:
        The base -- how many current units the next (single) unit contains,
        e.g., for an hour in context of days::

            CustomBaseUnit("hour", 24)

    :param overflow_after:     Value upper limit.
    :param custom_short:       Use specified short form instead of first letter
                               of ``name`` when operating in double-value mode.
    :param collapsible_after:  Min threshold for double output to become a regular one.
    """

    name: str
    in_next: int = None
    overflow_after: int = None
    custom_short: str = None
    collapsible_after: int = None

    def __post_init__(self):
        if not self.in_next and not self.overflow_after:
            raise ValueError("Either in_next or overflow_after is required.")
        if self.in_next and self.overflow_after:
            raise ValueError("Mutually exclusive attributes: in_next, overflow_after.")


class _TimeDeltaFormatterRegistry:
    """
    Simple tdf_registry for storing formatters and selecting
    the suitable one by max output length.
    """

    def __init__(self):
        self._formatters: t.Dict[int, DynamicBaseFormatter] = dict()

    def register(self, *formatters: DynamicBaseFormatter):
        for formatter in formatters:
            self._formatters[formatter.max_len] = formatter

    def find_matching(self, max_len: int) -> DynamicBaseFormatter | None:
        exact_match = self.get_by_max_len(max_len)
        if exact_match is not None:
            return exact_match

        suitable_max_len_list = sorted(
            [key for key in self._formatters.keys() if key <= max_len],
            key=lambda k: k,
            reverse=True,
        )
        if len(suitable_max_len_list) == 0:
            return None
        return self.get_by_max_len(suitable_max_len_list[0])

    def get_by_max_len(self, max_len: int) -> DynamicBaseFormatter | None:
        return self._formatters.get(max_len, None)

    def get_shortest(self) -> _TimeDeltaFormatterRegistry | None:
        return self._formatters.get(min(self._formatters.keys() or [None]))

    def get_longest(self) -> _TimeDeltaFormatterRegistry | None:
        return self._formatters.get(max(self._formatters.keys() or [None]))


tdf_registry = _TimeDeltaFormatterRegistry()


tdf_registry.register(
    DynamicBaseFormatter(
        [
            CustomBaseUnit("s", 60),
            CustomBaseUnit("m", 60),
            CustomBaseUnit("h", 24),
            CustomBaseUnit("d", overflow_after=99),
        ],
        allow_negative=False,
        allow_fractional=True,
        unit_separator=None,
        plural_suffix=None,
        overflow_msg="ERR",
    ),
    DynamicBaseFormatter(
        [
            CustomBaseUnit("s", 60),
            CustomBaseUnit("m", 60),
            CustomBaseUnit("h", 24),
            CustomBaseUnit("d", 30),
            CustomBaseUnit("M", 12),
            CustomBaseUnit("y", overflow_after=99),
        ],
        allow_negative=False,
        allow_fractional=True,
        unit_separator=" ",
        plural_suffix=None,
        overflow_msg="ERRO",
    ),
    DynamicBaseFormatter(
        [
            CustomBaseUnit("sec", 60),
            CustomBaseUnit("min", 60),
            CustomBaseUnit("hr", 24, collapsible_after=10),
            CustomBaseUnit("day", 30, collapsible_after=10),
            CustomBaseUnit("mon", 12),
            CustomBaseUnit("yr", overflow_after=99),
        ],
        allow_negative=False,
        allow_fractional=True,
        unit_separator=" ",
        plural_suffix=None,
    ),
    DynamicBaseFormatter(
        [
            CustomBaseUnit("sec", 60),
            CustomBaseUnit("min", 60, custom_short="min"),
            CustomBaseUnit("hour", 24, collapsible_after=24),
            CustomBaseUnit("day", 30, collapsible_after=10),
            CustomBaseUnit("month", 12),
            CustomBaseUnit("year", overflow_after=999),
        ],
        allow_negative=True,
        allow_fractional=True,
        unit_separator=" ",
        plural_suffix="s",
    ),
)



# -----------------------------------------------------------------------------
#  pytermor [ANSI formatted terminal output toolset]
#  (c) 2022-2023. A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
"""
.. testsetup:: *

    from pytermor.color import *

"""
from __future__ import annotations

import dataclasses
import math
import re
import typing as t
from abc import abstractmethod, ABC

from .ansi import (
    SequenceSGR,
    NOOP_SEQ,
    HI_COLORS,
    BG_HI_COLORS,
    make_color_256,
    make_color_rgb,
)
from .common import LogicError, CDT

CT = t.TypeVar("CT", bound="IColor")
""" 
Any non-abstract `IColor` type.

:meta public:
 """


class _ColorRegistry(t.Generic[CT], t.Sized):
    _TOKEN_SEPARATOR = "-"
    _QUERY_SPLIT_REGEX = re.compile(r"[\W_]+|(?<=[a-z])(?=[A-Z0-9])")

    def __init__(self):
        self._map: t.Dict[t.Tuple[str], CT] = {}

    def register(self, color: CT, name: str):
        primary_tokens = tuple(self._QUERY_SPLIT_REGEX.split(name))
        self._register_pair(color, primary_tokens)

        if not isinstance(color, ColorRGB):
            return
        for variation in color.variations.values():
            variation_tokens: t.Tuple[str, ...] = (
                *primary_tokens,
                *(self._QUERY_SPLIT_REGEX.split(variation.name)),
            )
            self._register_pair(variation, variation_tokens)

    def _register_pair(self, color: CT, tokens: t.Tuple[str, ...]):
        if tokens not in self._map.keys():
            self._map[tokens] = color
            return

        existing_color = self._map.get(tokens)
        if color.hex_value == existing_color.hex_value:
            return  # skipping the duplicate with the same name and value
        raise ColorNameConflictError(tokens, existing_color, color)

    def resolve(self, name: str) -> CT:
        query_tokens = (*(qt.lower() for qt in self._QUERY_SPLIT_REGEX.split(name)),)
        if color := self._map.get(query_tokens, None):
            return color
        raise LookupError(f"Color '{name}' does not exist")

    def __len__(self) -> int:
        return len(self._map)


class _ColorIndex(t.Generic[CT], t.Sized):
    def __init__(self):
        self._map: t.Dict[int, _ColorChannels[CT]] = {}

    def add(self, color: CT, code: int = None):
        if code is None:
            code = len(self._map)
        if code not in self._map.keys():
            self._map[code] = _ColorChannels(color)
            return

        existing_color = self._map.get(code).color
        if color.hex_value == existing_color.hex_value:
            return  # skipping the duplicate with the same code and value
        raise ColorCodeConflictError(code, existing_color, color)

    def get(self, code: int) -> CT:
        if channels := self._map.get(code, None):
            return channels.color
        raise KeyError(f"Color #{code} does not exist")

    def __len__(self) -> int:
        return len(self._map)

    @property
    def values(self) -> t.Iterable[_ColorChannels[CT]]:
        return self._map.values()


class _ColorChannels(t.Generic[CT]):
    def __init__(self, color: CT):
        self.color: CT = color
        self.r, self.g, self.b = self.color.to_rgb()


@dataclasses.dataclass(frozen=True)
class ApxResult(t.Generic[CT]):
    """
    Approximation result.
    """

    color: CT
    """ Found `IColor` instance. """
    distance: int
    """ Squared sRGB distance from this instance to the approximation target. """

    @property
    def distance_real(self) -> float:
        """
        Actual distance from instance to target:

            :math:`distance_{real} = \\sqrt{distance}`
        """
        return math.sqrt(self.distance)


class IColor(ABC):
    """
    Abstract superclass for other ``Colors``.

    :meta private:
    """

    # class vars #
    _registry: _ColorRegistry
    _index: _ColorIndex
    _approx_cache: t.Dict[int, CT]

    @classmethod
    def __new__(cls, *args, **kwargs):
        # @TODO find a way to provide a correct constructor signature
        #       for Color256 etc. instead of this one (for sphinx)
        # fmt: off
        if not hasattr(cls, "_registry"):     cls._registry = _ColorRegistry[cls]()
        if not hasattr(cls, "_index"):        cls._index = _ColorIndex[cls]()
        if not hasattr(cls, "_approx_cache"): cls._approx_cache = dict()
        return super().__new__(cls)
        # fmt: on

    def __init__(self, hex_value: int, name: str = None):
        if hex_value < 0 or hex_value > 0xFFFFFF:
            raise ValueError(f"Out of bounds hex value {hex_value:06X}, should be: 0x0 <= hex_value <= 0xFFFFFF")
        self._hex_value: int = hex_value
        self._name: str | None = name

    def _post_init(
        self: CT, code: int | None, register: bool, index: bool, aliases: t.List[str]
    ):
        if register:
            self._register_names(aliases)
        if index:
            self._index.add(self, code)

    def _register_names(self: CT, aliases: t.List[str] = None):
        if not self.name:
            return
        self._registry.register(self, self.name)

        if not aliases:
            return
        for alias in aliases:
            self._registry.register(self, alias)

    def _make_variations(self: CT, variation_map: t.Dict[int, str] = None):
        if not variation_map:
            return

        # @TODO rewrite this part later. don't like that entity relationship attributes
        #       are being set in some indistinct method, not in the constructor.
        for vari_hex_value, vari_name in variation_map.items():
            variation = type(self)(
                hex_value=vari_hex_value, name=vari_name, register=False, index=True
            )  # registration will be made by registry itself
            variation._base = self
            self._variations[vari_name] = variation

    def __eq__(self, other) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return self._hex_value == other._hex_value

    def format_value(self, prefix: str = "0x") -> str:
        """
        Format color value as "0xFFFFFF".

        :param prefix: Can be customized.
        """
        return f"{prefix:s}{self._hex_value:06X}"

    @property
    def hex_value(self) -> int:
        """Color value, e.g. 0x3AEB0C."""
        return self._hex_value

    @property
    def name(self) -> str | None:
        """Color name, e.g. "navy-blue"."""
        return self._name

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}[{self.repr_attrs()}]>"

    @abstractmethod
    def repr_attrs(self, verbose: bool = True) -> str:
        raise NotImplementedError

    def to_hsv(self) -> t.Tuple[float, float, float]:
        """
        Wrapper around `hex_to_hsv()` for concrete instance.

        :see: hex_to_hsv() for the details
        """
        return hex_to_hsv(self._hex_value)

    def to_rgb(self) -> t.Tuple[int, int, int]:
        """
        Wrapper around `to_rgb()` for concrete instance.

        :see: to_rgb() for the details
        """
        return hex_to_rgb(self._hex_value)

    @abstractmethod
    def to_sgr(self, bg: bool, upper_bound: t.Type[IColor] = None) -> SequenceSGR:
        """
        Make an `SGR sequence<SequenceSGR>` out of `IColor`. Used by `SgrRenderer`.

        :param bg: Set to *True* if required SGR should change the background color, or
                   *False* for the foreground (=text) color.
        :param upper_bound: Required result `IColor` type upper boundary, i.e., the
                            maximum acceptable color class, which will be the basis for
                            SGR being made. See `Color256.to_sgr()` for the details.
        """
        raise NotImplementedError

    @abstractmethod
    def to_tmux(self, bg: bool) -> str:
        """
        Make a tmux markup directive, which will change the output color to
        this color's value (after tmux processes and prints it). Used by `TmuxRenderer`.

        :param bg: Set to *True* if required tmux directive should change the background
                   color, or *False* for the foreground (=text) color.
        """
        raise NotImplementedError

    @classmethod
    def resolve(cls, name: str) -> CT:
        """
        Case-insensitive search through registry contents.

        :see: `resolve_color()` for the details
        :param name:  `IColor` name to search for.
        """
        if not hasattr(cls, "_registry"):
            raise LogicError(
                "Registry is empty. Did you call an abstract class' method?"
            )
        return cls._registry.resolve(name)

    @classmethod
    def find_closest(cls, hex_value: int) -> CT:
        """
        Search and return nearest to ``hex_value`` color instance.

        :see: `color.find_closest()` for the details
        :param hex_value: Target RGB value.
        """
        if not hasattr(cls, "_index"):
            raise LogicError("Index is empty. Did you call an abstract class' method?")

        if hex_value in cls._approx_cache.keys():
            return cls._approx_cache.get(hex_value)

        closest = cls._find_neighbours(hex_value)[0].color
        cls._approx_cache[hex_value] = closest
        return closest

    @classmethod
    def approximate(cls, hex_value: int, max_results: int = 1) -> t.List[ApxResult[CT]]:
        """
        Search for the colors nearest to ``hex_value`` and return the first ``max_results``.

        :see: `color.approximate()` for the details
        :param hex_value:   Target RGB value.
        :param max_results: Result limit.
        """
        if not hasattr(cls, "_index"):
            raise LogicError("Index is empty. Did you call an abstract class' method?")
        return cls._find_neighbours(hex_value)[:max_results]

    @classmethod
    def _find_neighbours(cls: t.Type[CT], hex_value: int) -> t.List[ApxResult[CT]]:
        """
        Iterate the registered colors table and compute the squared euclidean distance
        from argument to each color of the palette. Sort the results and return them.

        **sRGB euclidean distance**
            https://en.wikipedia.org/wiki/Color_difference#sRGB
            https://stackoverflow.com/a/35114586/5834973

        :param hex_value: Target RGB value.
        """
        input_r, input_g, input_b = hex_to_rgb(hex_value)
        result: t.List[ApxResult[CT]] = list()

        for channels in cls._index.values:
            distance_sq: int = (
                pow(channels.r - input_r, 2)
                + pow(channels.g - input_g, 2)
                + pow(channels.b - input_b, 2)
            )
            result.append(ApxResult(channels.color, distance_sq))
        return sorted(result, key=lambda r: r.distance)


class Color16(IColor):
    """
    Variant of a `IColor` operating within the most basic color set
    -- **Xterm-16**. Represents basic color-setting SGRs with primary codes
    30-37, 40-47, 90-97 and 100-107 (see `guide.presets.color16`).
    
    .. note ::

        Arguments ``register``, ``index`` and ``aliases``
        are *kwonly*-type args.

    :param hex_value: Color RGB value, e.g. 0x800000.
    :param code_fg:   Int code for a foreground color setup, e.g. 30.
    :param code_bg:   Int code for a background color setup. e.g. 40.
    :param name:      Name of the color, e.g. "red".
    :param register:  If *True*, add color to registry for resolving by name.
    :param index:     If *True*, add color to approximation index.
    :param aliases:   Alternative color names (used in `resolve_color()`).
    """

    def __init__(
        self,
        hex_value: int,
        code_fg: int,
        code_bg: int,
        name: str = None,
        *,
        register: bool = False,
        index: bool = False,
        aliases: t.List[str] = None,
    ):
        super().__init__(hex_value, name)
        self._code_fg: int = code_fg
        self._code_bg: int = code_bg
        self._post_init(self._code_fg, register, index, aliases)

    @property
    def code_fg(self) -> int:
        """Int code for a foreground color setup, e.g. 30."""
        return self._code_fg

    @property
    def code_bg(self) -> int:
        """Int code for a background color setup. e.g. 40."""
        return self._code_bg

    @classmethod
    def get_by_code(cls, code: int) -> Color16:
        """
        Get a `Color16` instance with specified code. Only *foreground* (=text) colors
        are indexed, therefore it is impossible to look up for a `Color16` with
        given background color.

        :param code:      Foreground integer code to look up for (see
                          `guide.presets.color16`).
        :raises KeyError: If no color with specified code is found.
        """
        return cls._index.get(code)

    def __eq__(self, other) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return (
            self._hex_value == other._hex_value
            and self._code_bg == other._code_bg
            and self._code_fg == other._code_fg
        )

    def repr_attrs(self, verbose: bool = True) -> str:
        # question mark after color value indicates that we cannot be 100% sure
        # about the exact value of xterm-16 colors, as they are configurable and
        # depend on terminal theme and settings. that's not the case for xterm-256,
        # though -- it's almost guaranteed to have the same color nearly everywhere.
        # the exceptions are rare and include color mapping at low level, e.g.,
        # ``tmux`` with specifically configured terminal capability overrides.
        # that's not something that you'd expect from a regular user, anyway.
        code = f"#{self._code_fg}"
        value = f"{self.format_value('')}?"

        params = [self._name]
        if verbose:
            params = [code, value] + params
        return ",".join(str(s) for s in filter(None, params))

    def to_sgr(self, bg: bool, upper_bound: t.Type[IColor] = None) -> SequenceSGR:
        if bg:
            return SequenceSGR(self._code_bg)
        return SequenceSGR(self._code_fg)

    def to_tmux(self, bg: bool) -> str:
        if self._name is None:
            raise LogicError("Translation to tmux format failed: color name required")
        code = self._code_bg if bg else self._code_fg
        is_hi = code in HI_COLORS or code in BG_HI_COLORS
        tmux_name = ("bright" if is_hi else "") + self._name.lower().replace("hi-", "")
        return tmux_name


class Color256(IColor):
    """
    Variant of a `IColor` operating within relatively modern **Xterm-256**
    indexed color table. Represents SGR complex codes ``38;5;*`` and ``48;5;*``
    (see `guide.presets.color256`).

    .. note ::

        Arguments ``register``, ``index``, ``aliases`` and ``color16_equiv``
        are *kwonly*-type args.

    :param hex_value: Color RGB value, e.g. 0x5f0000.
    :param code:      Int code for a color setup, e.g. 52.
    :param name:      Name of the color, e.g. "dark-red".
    :param register:  If *True*, add color to registry for resolving by name.
    :param index:     If *True*, add color to approximation index.
    :param aliases:   Alternative color names (used in `resolve_color()`).
    :param color16_equiv:
                      `Color16` counterpart (applies only to codes 0-15).
    """

    def __init__(
        self,
        hex_value: int,
        code: int,
        name: str = None,
        *,
        register: bool = False,
        index: bool = False,
        aliases: t.List[str] = None,
        color16_equiv: Color16 = None,
    ):
        super().__init__(hex_value, name)
        self._code: int | None = code
        self._color16_equiv: Color16 | None = None
        if color16_equiv:
            self._color16_equiv = Color16.get_by_code(color16_equiv.code_fg)
        self._post_init(self._code, register, index, aliases)

    def to_sgr(self, bg: bool, upper_bound: t.Type[IColor] = None) -> SequenceSGR:
        """
        Make an `SGR sequence<SequenceSGR>` out of `IColor`. Used by `SgrRenderer`.

        Each `IColor` type represents one SGR type in the context of colors. For
        example, if ``upper_bound`` is set to `Color16`, the resulting SGR will always
        be one of 16-color index table, even if the original color was of different
        type -- it will be approximated just before the SGR assembling.

        The reason for this is the necessity to provide a similar look for all users
        with different terminal settings/ capabilities. When the library sees that
        user's output device supports 256 colors only, it cannot assemble True Color
        SGRs, because they will be ignored (if we are lucky), or displayed in a
        glitchy way, or mess up the output completely. The good news is that the
        process is automatic and in most cases the library will manage the
        transformations by itself. If it's not the case, the developer can correct the
        behaviour by overriding the renderers' output mode. See `SgrRenderer` and
        `OutputMode` docs.

        :param bg: Set to *True* if required SGR should change the background color, or
                   *False* for the foreground (=text) color.
        :param upper_bound: Required result `IColor` type upper boundary, i.e., the
                            maximum acceptable color class, which will be the basis for
                            SGR being made.
        """
        if upper_bound is ColorRGB:
            return make_color_rgb(*self.to_rgb(), bg)

        if upper_bound is Color256 or upper_bound is None:
            return make_color_256(self._code, bg)

        if self._color16_equiv:
            return self._color16_equiv.to_sgr(bg, upper_bound)

        return Color16.find_closest(self.hex_value).to_sgr(bg, upper_bound)

    def to_tmux(self, bg: bool) -> str:
        return f"colour{self._code}"

    @property
    def code(self) -> int:
        """Int code for a color setup, e.g. 52."""
        return self._code

    @classmethod
    def get_by_code(cls, code: int) -> Color256:
        """
        Get a `Color256` instance with specified code (=position in the index).

        :param code:      Color code to look up for (see `guide.presets.color256`).
        :raises KeyError: If no color with specified code is found.
        """
        return cls._index.get(code)

    def __eq__(self, other) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return self._hex_value == other._hex_value and self._code == other._code

    def repr_attrs(self, verbose: bool = True) -> str:
        code = f"X{self._code}"
        value = self.format_value("")

        params = [code+f"[{value}]"]
        if verbose:
            params = [code, value, self._name]
        return ",".join(str(s) for s in filter(None, params))


class ColorRGB(IColor):
    """
    Variant of a `IColor` operating within RGB color space. Presets include
    `es7s named colors <guide.es7s-colors>`, a unique collection of colors
    compiled from several known sources after careful selection. However,
    it's not limited to aforementioned color list and can be easily extended.

    .. note ::

        Arguments ``register``, ``index``, ``aliases`` and ``variation_map``
        are *kwonly*-type args.


    :param hex_value: Color RGB value, e.g. 0x73a9c2.
    :param name:      Name of the color, e.g. "moonstone-blue".
    :param register:  If *True*, add color to registry for resolving by name.
    :param index:     If *True*, add color to approximation index.
    :param aliases:   Alternative color names (used in `resolve_color()`).
    :param variation_map: Mapping {*int*: *str*}, where keys are hex values,
                          and values are variation names.
    """

    def __init__(
        self,
        hex_value: int,
        name: str = None,
        *,
        register: bool = False,
        index: bool = False,
        aliases: t.List[str] = None,
        variation_map: t.Dict[int, str] = None,
    ):
        super().__init__(hex_value, name)
        self._base: CT | None = None
        self._variations: t.Dict[str, CT] = {}
        self._make_variations(variation_map)
        self._post_init(None, register, index, aliases)

    def to_sgr(self, bg: bool, upper_bound: t.Type[IColor] = None) -> SequenceSGR:
        if upper_bound is ColorRGB or upper_bound is None:
            return make_color_rgb(*self.to_rgb(), bg)

        return Color256.find_closest(self._hex_value).to_sgr(bg, upper_bound)

    def to_tmux(self, bg: bool) -> str:  # rgb hex format should be lower cased!
        return self.format_value("#").lower()

    def __eq__(self, other) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return self._hex_value == other._hex_value

    def repr_attrs(self, verbose: bool = True) -> str:
        value = self.format_value("")

        params = [value]
        if verbose:
            params += [self.name]
        return ",".join(str(s) for s in filter(None, params))

    @property
    def base(self) -> CT | None:
        """Parent color for color variations. Empty for regular colors."""
        return self._base

    @property
    def variations(self) -> t.Dict[str, CT]:
        """
        List of color variations. *Variation* of a color is a similar color with
        almost the same name, but with differing suffix. The main idea of variations
        is to provide a basis for fuzzy searching, which will return several results
        for one query; i.e., when the query matches a color with variations, the whole
        color family can be considered a match, which should increase searching speed.
        """
        return self._variations


class _NoopColor(IColor):
    def __init__(self):
        super().__init__(0)

    def __bool__(self) -> bool:
        return False

    def to_sgr(self, bg: bool, upper_bound: t.Type[IColor] = None) -> SequenceSGR:
        return NOOP_SEQ

    def to_tmux(self, bg: bool) -> str:
        return ""

    @property
    def hex_value(self) -> int:
        raise LogicError("No color for NO-OP instance")

    def format_value(self, prefix: str = "0x") -> str:
        return (prefix if "=" in prefix else "") + "NOP"

    def repr_attrs(self, verbose: bool = True) -> str:
        return self.format_value()


NOOP_COLOR = _NoopColor()
"""
Special `IColor` instance always rendering into empty string.
"""


def resolve_color(subject: CDT, color_type: t.Type[CT] = None) -> CT:
    """
    Case-insensitive search through registry contents. Search is performed for
    `IColor` instance named as specified in ``subject`` argument, and of specified
    ``color_type``, or for any type if argument is omitted: first it will be performed
    in the registry of `Color16` class, then -- in `Color256`, and, if previous two
    were unsuccessful, in the largest `ColorRGB` registry. Therefore, the return value
    could be any of these types:

        >>> resolve_color('red')
        <Color16[#31,800000?,red]>

    If ``color_type`` is `ColorRGB` or if it is omitted, there is one more way
    to specify a color: in form of a hexadecimal value "#RRGGBB" (or in short form, as
    "#RGB"), as well as just use an *int* in [0x0; 0xFFFFFF] range. In this case no
    actual searching is performed, and a new nameless instance of `ColorRGB` is
    created and returned.

        >>> resolve_color("#333")
        <ColorRGB[333333]>
        >>> resolve_color(0xfafef0)
        <ColorRGB[FAFEF0]>

    Color names are stored in registries as tokens, which allows to use any form of
    input and get the correct result regardless. The only requirement is to split the
    words in any matter, so that tokenizer could distinguish the words from each other:

        >>> resolve_color('deep-sky-blue-7')
        <Color256[X23,005F5F,deep-sky-blue-7]>
        >>> resolve_color('DEEP_SKY_BLUE_7')
        <Color256[X23,005F5F,deep-sky-blue-7]>
        >>> resolve_color('DeepSkyBlue7')
        <Color256[X23,005F5F,deep-sky-blue-7]>

        >>> resolve_color('deepskyblue7')
        Traceback (most recent call last):
        LookupError: Color 'deepskyblue7' was not found in any of registries

    :param str|int subject:
            `IColor` name or hex value to search for. See `CDT`.
    :param color_type:   Target color type (`Color16`, `Color256` or `ColorRGB`).
    :raises LookupError: If nothing was found in either of registries.
    :return:             `IColor` instance with specified name or value.
    """
    def subject_as_hex():
        nonlocal subject
        if isinstance(subject, int):
            return subject
        elif re.fullmatch(r"#[\da-f]{3}([\da-f]{3})?", subject, flags=re.IGNORECASE):
            subject = subject[1:]
            if len(subject) == 3:
                # 3-digit hex notation, basically #RGB -> #RRGGBB
                # https://www.quackit.com/css/color/values/css_hex_color_notation_3_digits.cfm
                subject = "".join(2*c for c in subject)
            return int(subject, 16)
        return None

    if (hex_value := subject_as_hex()) is not None:
        if color_type is None or color_type == ColorRGB:
            return ColorRGB(hex_value)

    color_types = [color_type] if color_type else [Color16, Color256, ColorRGB]
    for color_cls in color_types:
        try:
            return color_cls.resolve(str(subject))
        except LookupError:
            continue
    raise LookupError(f"Color '{subject}' was not found in any of registries")


def find_closest(hex_value: int, color_type: t.Type[CT] = None) -> CT:
    """
    Search and return nearest to ``hex_value`` instance of specified ``color_type``.
    If `color_type` is omitted, search for the closest `Color256` element.

    Method is useful for finding applicable color alternatives if user's
    terminal is incapable of operating in more advanced mode. Usually it is
    done by the library automatically and transparently for both the developer
    and the end-user.

    .. note ::

        This method caches the results, i.e., the same search query will from then
        onward result in the same return value without the necessity of iterating
        through the color index. If that's not applicable, use similar method
        `approximate()`, which is unaware of caching mechanism altogether.

    :param hex_value:   Target color RGB value.
    :param color_type:  Target color type (`Color16`, `Color256` or `ColorRGB`).
    :return:            Nearest to ``hex_value`` color instance of specified type.
    """
    return (color_type or Color256).find_closest(hex_value)


def approximate(
    hex_value: int, color_type: t.Type[CT] = None, max_results: int = 1
) -> t.List[ApxResult[CT]]:
    """
    Search for nearest to ``hex_value`` colors of specified ``color_type`` and
    return the first ``max_results`` of them. If `color_type` is omitted, search
    for the closest `Color256` elements. This method is similar to the
    `find_closest()`, although they differ in some aspects:

        - `approximate()` can return more than one result;
        - `approximate()` returns not just a `IColor` instance(s), but also a
          number equal to squared distance to the target color for each of them;
        - `find_closest()` caches the results, while `approximate()` ignores
          the cache completely.

    :param hex_value:    Target color RGB value.
    :param color_type:   Target color type (`Color16`, `Color256` or `ColorRGB`).
    :param max_results:  Return no more than ``max_results`` items.
    :return: Pairs of closest `IColor` instance(s) found with their distances
             to the target color, sorted by distance descending, i.e., element
             at index 0 is the closest color found, paired with its distance
             to the target; element with index 1 is second-closest color
             (if any) and corresponding distance value, etc.
    """
    return (color_type or Color256).approximate(hex_value, max_results)


def hex_to_rgb(hex_value: int) -> t.Tuple[int, int, int]:
    """
    Transforms ``hex_value`` in *int* format into a tuple of three
    integers corresponding to **red**, **blue** and **green** channel value
    respectively. Values are within [0; 255] range.

        >>> hex_to_rgb(0x80ff80)
        (128, 255, 128)

    :param hex_value: RGB value.
    :returns: R, G, B channel values correspondingly.
    """
    if not isinstance(hex_value, int):
        raise TypeError(f"Argument type should be 'int', got: {type(hex_value)}")

    return ((hex_value & 0xFF0000) >> 16, (hex_value & 0xFF00) >> 8, (hex_value & 0xFF))


def rgb_to_hex(r: int, g: int, b: int) -> int:
    """
    Transforms RGB value in a three-integers form ([0; 255], [0; 255], [0; 255])
    to an one-integer form.

        >>> hex(rgb_to_hex(0, 128, 0))
        '0x8000'

    :param r: value of red channel.
    :param g: value of green channel.
    :param b: value of blue channel.
    :return: RGB value.
    """
    return (r << 16) + (g << 8) + b


def hsv_to_rgb(h: float, s: float, v: float) -> t.Tuple[int, int, int]:
    """
    Transforms HSV value in three-floats form (where 0 <= h < 360, 0 <= s <= 1,
    and 0 <= v <= 1) into RGB three-integer form ([0; 255], [0; 255], [0; 255]).

        >>> hsv_to_rgb(270, 2/3, 0.75)
        (128, 64, 192)

    :param h: hue channel value.
    :param s: saturation channel value.
    :param v: value channel value.
    :return: R, G, B channel values correspondingly.
    """
    h = 0.0 if h == 360.0 else h / 60.0
    fract = h - math.floor(h)

    p = v * (1.0 - s)
    q = v * (1.0 - s * fract)
    t = v * (1.0 - s * (1.0 - fract))

    if 0.0 <= h < 1.0:
        r, g, b = v, t, p
    elif 1.0 <= h < 2.0:
        r, g, b = q, v, p
    elif 2.0 <= h < 3.0:
        r, g, b = p, v, t
    elif 3.0 <= h < 4.0:
        r, g, b = p, q, v
    elif 4.0 <= h < 5.0:
        r, g, b = t, p, v
    elif 5.0 <= h < 6.0:
        r, g, b = v, p, q
    else:
        r, g, b = 0, 0, 0

    r, g, b = (math.ceil(255 * c) for c in (r, g, b))
    return r, g, b


def rgb_to_hsv(r: int, g: int, b: int) -> t.Tuple[float, float, float]:
    """
    Transforms RGB value in a three-integers form ([0; 255], [0; 255], [0; 255]) to an
    HSV in three-floats form such as (0 <= h < 360, 0 <= s <= 1, and 0 <= v <= 1).

        >>> rgb_to_hsv(0, 0, 255)
        (240.0, 1.0, 1.0)

    :param r: value of red channel.
    :param g: value of green channel.
    :param b: value of blue channel.
    :returns: H, S, V channel values correspondingly.
    """
    # fmt: off
    # https://en.wikipedia.org/wiki/HSL_and_HSV#From_RGB

    rn, gn, bn = r / 255, g / 255, b / 255
    vmax = max(rn, gn, bn)
    vmin = min(rn, gn, bn)
    c = vmax - vmin
    v = vmax

    h = 0
    if c == 0:     pass
    elif v == rn:  h = 60 * (0 + (gn - bn) / c)
    elif v == gn:  h = 60 * (2 + (bn - rn) / c)
    elif v == bn:  h = 60 * (4 + (rn - gn) / c)

    if v == 0:     s = 0
    else:          s = c / v

    if h < 0:      h += 360

    return h, s, v
    # fmt: on


def hex_to_hsv(hex_value: int) -> t.Tuple[float, float, float]:
    """
    Transforms ``hex_value`` in *int* form into a tuple of three numbers
    corresponding to **hue**, **saturation** and **value** channel values respectively.
    Hue is within [0, 359] range, both saturation and value are within [0; 1] range.

        >>> hex_to_hsv(0x999999)
        (0, 0.0, 0.6)

    :param hex_value: RGB value.
    :returns: H, S, V channel values correspondingly.
    """
    return rgb_to_hsv(*hex_to_rgb(hex_value))


def hsv_to_hex(h: float, s: float, v: float) -> int:
    """
    Transforms HSV value in three-floats form (where 0 <= h < 360, 0 <= s <= 1,
    and 0 <= v <= 1) into an one-integer form.

        >>> hex(hsv_to_hex(90, 0.5, 0.5))
        '0x608040'

    :param h: hue channel value.
    :param s: saturation channel value.
    :param v: value channel value.
    :return: RGB value.
    """
    return rgb_to_hex(*hsv_to_rgb(h, s, v))


class ColorNameConflictError(Exception):
    def __init__(self, tokens: t.Tuple[str], existing_color: CT, new_color: CT):
        msg = f"Color '{new_color.name}' -> {tokens} already exists"
        super().__init__(msg, [existing_color, new_color])


class ColorCodeConflictError(Exception):
    def __init__(self, code: int, existing_color: CT, new_color: CT):
        msg = f"Color #{code} already exists"
        super().__init__(msg, existing_color, new_color)
